"""
Module for Terraform Cloud API Endpoint: Admin Module Sharing.
"""

from .endpoint import TFCEndpoint

class TFCAdminModuleSharing(TFCEndpoint):
    """
    `Admin Module Sharing API Docs \
        <https://www.terraform.io/docs/cloud/api/admin/module-sharing.html#update-an-organization-39-s-module-consumers>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/admin/organizations"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return True

    def update(self, org_name, payload):
        """
        ``PATCH /admin/organizations/:name/module-consumers``

        `Admin Module Sharing API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/module-sharing.html#update-an-organization-39-s-module-consumers>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/admin/module-sharing.html#sample-payload>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}/module-consumers"
        return self._update(url, payload)



"""
Module for testing the Terraform Cloud API Endpoint: Policy Sets.
"""

from .base import TestTFCBaseTestCase


class TestTFCPolicySetParams(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Policy Set Params.
    """

    _unittest_name = "pol-set-params"
    _endpoint_being_tested = "policy_set_params"

    def setUp(self):
        oauth_client = self._api.oauth_clients.create(\
            self._get_oauth_client_create_payload())["data"]
        self._oauth_client_id = oauth_client["id"]
        oauth_token_id = oauth_client["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a temp policy set to manipulate in the test, store the ID
        created_policy_set = self._api.policy_sets.create(\
            self._get_policy_set_create_payload(oauth_token_id))["data"]
        self._created_policy_set_id = created_policy_set["id"]

    def tearDown(self):
        # Destroy the workspace and policy we created
        self._api.policy_sets.destroy(self._created_policy_set_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_policy_sets(self):
        """
        Test the Policy Set Params API endpoints.
        """

        # List the params, confirm that there are none to start.
        all_params = self._api.policy_set_params.list(self._created_policy_set_id)["data"]
        self.assertEqual(len(all_params), 0)

        # Create a variable and confirm it was added to the policy set
        create_payload = {
            "data": {
                "type": "vars",
                "attributes": {
                    "key": "terrasnek",
                    "value": "unittest",
                    "category": "policy-set",
                    "sensitive": False
                }
            }
        }
        created_param = self._api.policy_set_params.create(\
            self._created_policy_set_id, create_payload)["data"]
        created_param_id = created_param["id"]
        all_params = self._api.policy_set_params.list(self._created_policy_set_id)["data"]
        self.assertEqual(len(all_params), 1)

        # Update the value and confirm the change
        value_to_update_to = "bar"
        update_payload = {
            "data": {
                "id": created_param_id,
                "attributes": {
                    "key": "foo",
                    "value": value_to_update_to,
                    "category": "policy-set",
                    "sensitive": False
                },
                "type": "vars"
            }
        }
        update_params = self._api.policy_set_params.update(\
            self._created_policy_set_id, created_param_id, update_payload)["data"]
        updated_value = update_params["attributes"]["value"]
        self.assertEqual(updated_value, value_to_update_to)

        # Delete the variable and confirm there are none left
        self._api.policy_set_params.destroy(self._created_policy_set_id, created_param_id)
        all_params = self._api.policy_set_params.list(self._created_policy_set_id)["data"]
        self.assertEqual(len(all_params), 0)



"""
Module for testing the Terraform Cloud API Endpoint: Admin Terraform Versions.
"""

from .base import TestTFCBaseTestCase

TF_RELEASE_URL = "https://releases.hashicorp.com/terraform/0.11.8/terraform_0.11.8_linux_amd64.zip"
TF_RELEASE_SHA = "84ccfb8e13b5fce63051294f787885b76a1fedef6bdbecf51c5e586c9e20c9b7"

class TestTFCAdminTerraformVersions(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Admin Terraform Versions.
    """

    _unittest_name = "tf-ver"
    _endpoint_being_tested = "admin_terraform_versions"

    def test_admin_terraform_versions(self):
        """
        Test the Admin Terraform Versions API endpoints.
        """

        # List all the TF versions, verify the response type
        all_tf_versions = self._api.admin_terraform_versions.list()["data"]
        self.assertEqual("terraform-versions", all_tf_versions[0]["type"])

        # Create a fake version and confirm it was created correctly
        version_to_create = "0.1.300"
        create_payload = {
            "data": {
                "type": "terraform-versions",
                "attributes": {
                    "version": version_to_create,
                    "url": TF_RELEASE_URL,
                    "sha": TF_RELEASE_SHA,
                    "official": False,
                    "enabled": True,
                    "beta": False
                }
            }
        }
        created_tf_version = self._api.admin_terraform_versions.create(create_payload)["data"]
        created_tf_version_id = created_tf_version["id"]
        self.assertEqual(version_to_create, created_tf_version["attributes"]["version"])

        # Update the newly created version to be a beta version, confirm the change
        beta_to_update_to = True
        update_payload = {
            "data": {
                "type": "terraform-versions",
                "attributes": {
                    "beta": beta_to_update_to
                }
            }
        }
        updated_tf_version = \
            self._api.admin_terraform_versions.update(created_tf_version_id, update_payload)["data"]
        self.assertEqual(beta_to_update_to, updated_tf_version["attributes"]["beta"])

        # Show the TF version, confirm the version number
        shown_tf_version = self._api.admin_terraform_versions.show(created_tf_version_id)["data"]
        self.assertEqual(version_to_create, shown_tf_version["attributes"]["version"])

        # Destroy the TF version, confirm it's gone
        self._api.admin_terraform_versions.destroy(created_tf_version_id)
        all_tf_versions = self._api.admin_terraform_versions.list()["data"]
        found_created_version = False
        for version in all_tf_versions:
            version_id = version["id"]
            if version_id == created_tf_version_id:
                found_created_version = True
                break
        self.assertFalse(found_created_version)



"""
Module for Terraform Cloud API Endpoint: Workspaces.
"""

from .endpoint import TFCEndpoint
from ._constants import MAX_PAGE_SIZE

class TFCWorkspaces(TFCEndpoint):
    """
    `Workspaces API Docs \
        <https://www.terraform.io/docs/cloud/api/workspaces.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._ws_api_v2_base_url = f"{self._api_v2_base_url}/workspaces"
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations/{org_name}/workspaces"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST /organizations/:organization_name/workspaces``

        `Workspaces Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#create-a-workspace>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#sample-payload>`_
        """
        return self._create(self._org_api_v2_base_url, payload)

    def destroy(self, workspace_id=None, workspace_name=None):
        """
        ``DELETE /organizations/:organization_name/workspaces/:name``
        ``DELETE /workspaces/:workspace_id``

        `Workspaces Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#delete-a-workspace>`_
        """
        if workspace_name is not None:
            url = f"{self._org_api_v2_base_url}/{workspace_name}"
        elif workspace_id is not None:
            url = f"{self._ws_api_v2_base_url}/{workspace_id}"
        else:
            self._logger.error("Arguments workspace_name or workspace_id must be defined")

        return self._destroy(url)

    def force_unlock(self, workspace_id):
        """
        ``POST /workspaces/:workspace_id/actions/force-unlock``

        `Workspaces Force Unlock API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#force-unlock-a-workspace>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/actions/force-unlock"
        return self._post(url)

    def lock(self, workspace_id, payload):
        """
        ``POST /workspaces/:workspace_id/actions/lock``

        `Workspaces Lock API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#lock-a-workspace>`_

        `Lock Sample Payload \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#sample-payload-2>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/actions/lock"
        return self._post(url, data=payload)

    def list(self, page=None, page_size=None):
        """
        ``GET /organizations/:organization_name/workspaces``

        `Workspaces List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#list-workspaces>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#query-parameters>`_):
            - ``page`` (Optional)
            - ``page_size`` (Optional)
        """
        return self._list(self._org_api_v2_base_url, page=page, page_size=page_size)

    def list_all(self):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every workspace in an organization.

        Returns an array of objects.
        """
        current_page_number = 1
        workspaces_resp = \
            self.list(page=current_page_number, page_size=MAX_PAGE_SIZE)
        total_pages = workspaces_resp["meta"]["pagination"]["total-pages"]

        workspaces = []
        while current_page_number <= total_pages:
            workspaces_resp = \
                self._list(self._org_api_v2_base_url, \
                    page=current_page_number, page_size=MAX_PAGE_SIZE)
            workspaces += workspaces_resp["data"]
            current_page_number += 1

        return workspaces

    def show(self, workspace_name=None, workspace_id=None):
        """
        ``GET /organizations/:organization_name/workspaces/:name``
        ``GET /workspaces/:workspace_id``

        `Workspaces Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#show-workspace>`_
        """
        if workspace_name is not None:
            url = f"{self._org_api_v2_base_url}/{workspace_name}"
        elif workspace_id is not None:
            url = f"{self._ws_api_v2_base_url}/{workspace_id}"
        else:
            self._logger.error("Arguments workspace_name or workspace_id must be defined")

        return self._show(url)

    def unlock(self, workspace_id):
        """
        ``POST /workspaces/:workspace_id/actions/unlock``

        `Workspaces Unlock API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#unlock-a-workspace>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/actions/unlock"
        return self._post(url)

    def update(self, payload, workspace_name=None, workspace_id=None):
        """
        ``PATCH /organizations/:organization_name/workspaces/:name``
        ``PATCH /workspaces/:workspace_id``

        `Workspaces Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#update-a-workspace>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#sample-payload-1>`_
        """
        if workspace_name is not None:
            url = f"{self._org_api_v2_base_url}/{workspace_name}"
        elif workspace_id is not None:
            url = f"{self._ws_api_v2_base_url}/{workspace_id}"
        else:
            self._logger.error("Arguments workspace_name or workspace_id must be defined")
        return self._update(url, payload)

    def assign_ssh_key(self, workspace_id, payload):
        """
        ``PATCH /workspaces/:workspace_id/relationships/ssh-key``

        `Workspaces Assign SSH Key API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#assign-an-ssh-key-to-a-workspace>`_

        `Assign Sample Payload \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#sample-payload-3>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/relationships/ssh-key"
        self._patch(url, data=payload)

    def unassign_ssh_key(self, workspace_id, payload):
        """
        ``PATCH /workspaces/:workspace_id/relationships/ssh-key``

        `Workspaces Unassign SSH Key API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#unassign-an-ssh-key-from-a-workspace>`_

        `Unassign Sample Payload \
            <https://www.terraform.io/docs/cloud/api/workspaces.html#sample-payload-4>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/relationships/ssh-key"
        self._patch(url, data=payload)



"""
Module for testing the Terraform Cloud API Endpoint: Workspace Variables.
"""

from .base import TestTFCBaseTestCase


class TestTFCWorkspaceVars(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Workspace Variables.
    """

    _unittest_name = "ws-vars"
    _endpoint_being_tested = "workspace_vars"

    def setUp(self):
        self._ws = self._api.workspaces.create(self._get_ws_without_vcs_create_payload())
        self._ws_id = self._ws["data"]["id"]
        self._ws_name = self._ws["data"]["attributes"]["name"]
        self._variable_test_key = "terrasnek_unittest_key"
        self._variable_test_value = "terrasnek_unittest_value"

    def tearDown(self):
        self._api.workspaces.destroy(
            workspace_name=self._ws["data"]["attributes"]["name"])

    def test_workspace_variables(self):
        """
        Test the Workspace Variables API endpoints.
        """

        create_variable_payload = self._get_variable_create_payload(
            self._variable_test_key, self._variable_test_value, self._ws_id)

        # Create the variable and assert the key/value pair is the same as defined
        variable = self._api.workspace_vars.create(self._ws_id, create_variable_payload)["data"]
        variable_value = variable["attributes"]["value"]
        self.assertEqual(self._variable_test_key,
                         variable["attributes"]["key"])
        self.assertEqual(self._variable_test_value, variable_value)

        # List the variables and make sure they match the updated payload
        original_variable_id = variable["id"]
        variables = self._api.workspace_vars.list(self._ws_id)["data"]
        self.assertEqual(original_variable_id, variables[0]["id"])

        # Now change the value of that variable
        updated_value = "changed"
        update_variable_payload = {
            "data": {
                "id": original_variable_id,
                "attributes": {
                    "key": self._variable_test_key,
                    "value": updated_value,
                    "category": "terraform",
                    "hcl": False,
                    "sensitive": False
                },
                "type": "vars"
            }
        }
        updated_variable = self._api.workspace_vars.update(
            self._ws_id, original_variable_id, update_variable_payload)["data"]
        self.assertEqual(
            updated_variable["attributes"]["value"], updated_value)

        # Delete the variable and confirm it's gone
        self._api.workspace_vars.destroy(self._ws_id, original_variable_id)
        variables = self._api.workspace_vars.list(self._ws_id)["data"]
        self.assertEqual(len(variables), 0)



"""
Module for Terraform Cloud API Endpoint: Cost Estimates.
"""

from .endpoint import TFCEndpoint

class TFCCostEstimates(TFCEndpoint):
    """
    `Cost Estimates API Docs \
        <https://www.terraform.io/docs/cloud/api/cost-estimates.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/cost-estimates"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def show(self, cost_est_id):
        """
        ``GET /cost-estimates/:id``

        `Cost Estimates Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/cost-estimates.html#show-a-cost-estimate>`_
        """
        url = f"{self._endpoint_base_url}/{cost_est_id}"
        return self._show(url)



"""
Module for testing the Terraform Cloud API Endpoint: Audit Trails.
"""

from .base import TestTFCBaseTestCase


class TestTFCAuditTrails(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Audit Trails.
    """

    _unittest_name = "audtrl"
    _endpoint_being_tested = "audit_trails"

    def setUp(self):
        # Create a user token with the API so that something shows in the audit logs for the org
        logged_in_user = self._api.account.show()["data"]
        logged_in_user_id = logged_in_user["id"]

        desc_to_update_to = self._unittest_random_name()
        create_payload = {
            "data": {
                "type": "authentication-tokens",
                "attributes": {
                    "description": desc_to_update_to
                }
            }
        }
        self._created_token = \
            self._api.user_tokens.create(logged_in_user_id, create_payload)["data"]
        self._created_token_id = self._created_token["id"]

    def tearDown(self):
        self._api.user_tokens.destroy(self._created_token_id)

    def test_audit_trails(self):
        """
        Test the Audit Trails API endpoints.
        """

        # List the audit trails. This deviates from the standard JSON spec
        # we use across the API, so we'll just assert no response, which
        # means it completed successfully for now.

        # TODO: audit trails aren't showing properly
        # print(self._test_org_name)
        # print(audit_trails)
        audit_trails = self._api.audit_trails.list()["data"]
        self.assertEqual(len(audit_trails), 0)

        all_audit_trails = self._api.audit_trails.list_all()
        self.assertEqual(len(all_audit_trails), 0)



"""
Module for testing the Terraform Cloud API Endpoint: Feature Sets.
"""

from .base import TestTFCBaseTestCase


class TestTFCFeatureSets(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Feature Sets.
    """

    _unittest_name = "fset"
    _endpoint_being_tested = "feature_sets"

    def test_feature_sets(self):
        """
        Test the Feature Sets API endpoints.
        """
        feature_sets = self._api.feature_sets.list()["data"]
        self.assertEqual("feature-sets", feature_sets[0]["type"])

        feature_sets_for_org = self._api.feature_sets.list_for_org()["data"]
        self.assertEqual("feature-sets", feature_sets_for_org[0]["type"])



"""
Module for testing the Terraform Cloud API Endpoint: Run Triggers.
"""

from .base import TestTFCBaseTestCase

RUN_TRIGGER_TYPE = "inbound"

class TestTFCRunTriggers(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Run Triggers.
    """

    _unittest_name = "run-trig"
    _endpoint_being_tested = "run_triggers"

    def setUp(self):
        self._source_ws = self._api.workspaces.create(self._get_ws_without_vcs_create_payload())
        self._source_ws_id = self._source_ws["data"]["id"]
        self._source_ws_name = self._source_ws["data"]["attributes"]["name"]

        self._target_ws = self._api.workspaces.create(self._get_ws_without_vcs_create_payload())
        self._target_ws_id = self._target_ws["data"]["id"]
        self._target_ws_name = self._target_ws["data"]["attributes"]["name"]

    def tearDown(self):
        self._api.workspaces.destroy(
            workspace_name=self._source_ws_name)
        self._api.workspaces.destroy(
            workspace_name=self._target_ws_name)

    def test_run_triggers(self):
        """
        Test the Run Triggers API endpoints.
        """

        test_filters = [
            {
                "keys": ["run-trigger", "type"],
                "value": RUN_TRIGGER_TYPE
            }
        ]

        # List the triggers and confirm there are none
        some_triggers = self._api.run_triggers.list(\
            self._target_ws_id, filters=test_filters)["data"]
        self.assertEqual(len(some_triggers), 0)

        # Create a trigger
        create_payload = {
            "data": {
                "relationships": {
                    "sourceable": {
                        "data": {
                            "id": self._source_ws_id,
                            "type": "workspaces"
                        }
                    }
                }
            }
        }
        created_trigger = self._api.run_triggers.create(\
            self._target_ws_id, create_payload)["data"]
        created_trigger_id = created_trigger["id"]

        # List the triggers again and confirm there is one
        some_triggers = self._api.run_triggers.list(\
            self._target_ws_id, filters=test_filters)["data"]
        self.assertEqual(len(some_triggers), 1)

        # List all triggers and confirm there is one
        all_triggers = self._api.run_triggers.list_all(self._target_ws_id, filters=test_filters)
        self.assertEqual(len(all_triggers), 1)

        # Show the run trigger by id, compare to our created ID
        shown_trigger = self._api.run_triggers.show(created_trigger_id)["data"]
        shown_trigger_id = shown_trigger["id"]
        self.assertEqual(shown_trigger_id, created_trigger_id)

        # Destroy the run trigger, confirm that we have zero run triggers again
        self._api.run_triggers.destroy(created_trigger_id)
        some_triggers = self._api.run_triggers.list(
            self._target_ws_id, filters=test_filters)["data"]
        self.assertEqual(len(some_triggers), 0)



"""
Module for Terraform Cloud API Endpoint: Feature Sets.
"""

from .endpoint import TFCEndpoint

class TFCFeatureSets(TFCEndpoint):
    """
    `Feature Sets API Docs \
        <https://www.terraform.io/docs/cloud/api/feature-sets.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/feature-sets"
        self._org_base_url = \
            f"{self._api_v2_base_url}/organizations/{org_name}/feature-sets"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return True

    def terraform_enterprise_only(self):
        return False

    def list(self):
        """
        ``GET /feature-sets``

        `Feature Sets List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/feature-sets.html>`_
        """
        return self._list(self._endpoint_base_url)

    def list_for_org(self):
        """
        ``GET /organizations/:organization_name/feature-sets``

        `Feature Sets List for Orgs API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/feature-sets.html#list-feature-sets-for-organization>`_
        """
        return self._list(self._org_base_url)



"""
Module for Terraform Cloud API Endpoint: Policy Set Params.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCPolicySetParams(TFCEndpoint):
    """
    `Policy Set Params API Docs \
        <https://www.terraform.io/docs/cloud/api/policy-set-params.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/policy-sets"

    def required_entitlements(self):
        return [Entitlements.SENTINEL]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, policy_set_id, payload):
        """
        ``POST /policy-sets/:policy_set_id/parameters``

        `Policy Set Params Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-set-params.html#create-a-parameter>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policy-set-params.html#sample-payload>`_
        """
        url = f"{self._endpoint_base_url}/{policy_set_id}/parameters"
        return self._create(url, payload)

    def list(self, policy_set_id):
        """
        ``GET /policy-sets/:policy_set_id/parameters``

        `Policy Set Params List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-set-params.html#list-parameters>`_
        """
        url = f"{self._endpoint_base_url}/{policy_set_id}/parameters"
        return self._list(url)

    def update(self, policy_set_id, parameter_id, payload):
        """
        ``PATCH /policy-sets/:policy_set_id/parameters/:parameter_id``

        `Policy Set Params Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-set-params.html#update-parameters>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policy-set-params.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{policy_set_id}/parameters/{parameter_id}"
        return self._update(url, payload)

    def destroy(self, policy_set_id, parameter_id):
        """
        ``DELETE /policy-sets/:policy_set_id/parameters/:parameter_id``

        `Policy Set Params Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-set-params.html#delete-parameters>`_
        """
        url = f"{self._endpoint_base_url}/{policy_set_id}/parameters/{parameter_id}"
        return self._destroy(url)



"""
Module for Terraform Cloud API Endpoint: Audit Trails.
"""

from .endpoint import TFCEndpoint
from ._constants import MAX_PAGE_SIZE

class TFCAuditTrails(TFCEndpoint):
    """
    `Audit Trails API Docs \
        <https://www.terraform.io/docs/cloud/api/audit-trails.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._audit_trail_api_v2_base_url = f"{self._api_v2_base_url}/organization/audit-trail"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return True

    def terraform_enterprise_only(self):
        return False

    def list(self, since=None, page=None, page_size=None):
        """
        ``GET /organization/audit-trail``

        `Audit Trails List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/audit-trails.html#list-audit-trails>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/audit-trails.html#query-parameters>`_):
            - ``since`` (Optional)
            - ``page_size`` (Optional)
            - ``page`` (Optional)
        """

        return self._list(self._audit_trail_api_v2_base_url, \
            page=page, page_size=page_size, since=since)

    def list_all(self):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every audit trail in an organization.

        Returns an array of objects.
        """
        current_page_number = 1
        audit_trails_resp = \
            self.list(page=current_page_number, page_size=MAX_PAGE_SIZE)
        total_pages = audit_trails_resp["pagination"]["total_pages"]

        audit_trails = []
        while current_page_number <= total_pages:
            audit_trails_resp = \
                self._list(self._audit_trail_api_v2_base_url, \
                    page=current_page_number, page_size=MAX_PAGE_SIZE)
            audit_trails += audit_trails_resp["data"]
            current_page_number += 1

        return audit_trails



"""
Module for Terraform Cloud API Endpoint: Team Access.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCTeamAccess(TFCEndpoint):
    """
    `Team Access API Docs \
        <https://www.terraform.io/docs/cloud/api/team-access.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/team-workspaces"

    def required_entitlements(self):
        return [Entitlements.TEAMS]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def add_team_access(self, payload):
        """
        ``POST /team-workspaces``

        `Team Access Add API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/team-access.html#add-team-access-to-a-workspace>`_

        `Add Sample Payload \
            <https://www.terraform.io/docs/cloud/api/team-access.html#sample-payload>`_
        """
        return self._post(self._endpoint_base_url, data=payload)

    def list(self, filters=None):
        """
        ``GET /team-workspaces``

        `Team Access List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/team-access.html#list-team-access-to-a-workspace>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/team-access.html#query-parameters>`_):
            - ``filter[workspace][id]`` (Required)

        Example filter(s):

        .. code-block:: python

            filters = [
                {
                    "keys": ["workspace", "id"],
                    "value": "foo"
                }
            ]
        """
        return self._list(self._endpoint_base_url, filters=filters)

    def remove_team_access(self, access_id):
        """
        ``DELETE /team-workspaces/:id``

        `Team Access Remove API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/team-access.html#remove-team-access-to-a-workspace>`_
        """
        url = f"{self._endpoint_base_url}/{access_id}"
        return self._delete(url)

    def show(self, access_id):
        """
        ``GET /team-workspaces/:id``

        `Team Access Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/team-access.html#show-a-team-access-relationship>`_
        """
        url = f"{self._endpoint_base_url}/{access_id}"
        return self._show(url)

    def update(self, access_id, payload):
        """
        ``PATCH /team-workspaces/:id``

        `Team Access Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/team-access.html#update-team-access-to-a-workspace>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/team-access.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{access_id}"
        return self._update(url, payload)



"""
Module for testing the Terraform Cloud API Endpoint: User Tokens.
"""

from .base import TestTFCBaseTestCase


class TestTFCUserTokens(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: User Tokens.
    """

    _unittest_name = "user-tok"
    _endpoint_being_tested = "user_tokens"

    def test_user_tokens(self):
        """
        Test the User Tokens API endpoints.
        """

        logged_in_user = self._api.account.show()["data"]
        logged_in_user_id = logged_in_user["id"]

        desc_to_update_to = self._unittest_random_name()
        create_payload = {
            "data": {
                "type": "authentication-tokens",
                "attributes": {
                    "description": desc_to_update_to
                }
            }
        }
        created_token = self._api.user_tokens.create(logged_in_user_id, create_payload)["data"]
        created_token_id = created_token["id"]

        tokens = self._api.user_tokens.list(logged_in_user_id)["data"]
        found_token = False
        for token in tokens:
            if token["id"] == created_token_id:
                found_token = True
                break
        self.assertTrue(found_token)

        self._api.user_tokens.destroy(created_token_id)

        tokens = self._api.user_tokens.list(logged_in_user_id)["data"]
        found_token = False
        for token in tokens:
            if token["id"] == created_token_id:
                found_token = True
                break
        self.assertFalse(found_token)



"""
Module for Terraform Cloud API Endpoint: Admin Settings.
"""

from .endpoint import TFCEndpoint

class TFCAdminSettings(TFCEndpoint):
    """
    `Admin Settings API Docs \
        <https://www.terraform.io/docs/cloud/api/admin/settings.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/admin"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return True

    def list_general(self):
        """
        ``GET /api/v2/admin/general-settings``

        `Admin List General Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#list-general-settings>`_
        """
        url = f"{self._endpoint_base_url}/general-settings"
        return self._list(url)

    def update_general(self, data):
        """
        ``PATCH /api/v2/admin/general-settings``

        `Admin Update General Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#update-general-settings>`_
        """
        url = f"{self._endpoint_base_url}/general-settings"
        return self._patch(url, data=data)

    def list_cost_estimation(self):
        """
        ``GET /api/v2/admin/cost-estimation-settings``

        `Admin List Cost Estimation Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#list-cost-estimation-settings>`_
        """
        url = f"{self._endpoint_base_url}/cost-estimation-settings"
        return self._list(url)

    def update_cost_estimation(self, data):
        """
        ``PATCH /api/v2/admin/cost-estimation-settings``

        `Admin Update Cost Estimation Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#update-cost-estimation-settings>`_
        """
        url = f"{self._endpoint_base_url}/cost-estimation-settings"
        return self._patch(url, data)

    def list_saml(self):
        """
        ``GET /api/v2/admin/saml-settings``

        `Admin List SAML Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#list-saml-settings>`_
        """
        url = f"{self._endpoint_base_url}/saml-settings"
        return self._list(url)

    def update_saml(self, data):
        """
        ``PATCH /api/v2/admin/saml-settings``

        `Admin Update SAML Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#update-saml-settings>`_
        """
        url = f"{self._endpoint_base_url}/saml-settings"
        return self._patch(url, data)

    def revoke_previous_saml_idp_cert(self):
        """
        ``POST /api/v2/admin/saml-settings/actions/revoke-old-certificate``

        `Admin Revoke SAML IDP Cert API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#revoke-previous-saml-idp-certificate>`_
        """
        url = f"{self._endpoint_base_url}/saml-settings/actions/revoke-old-certificate"
        return self._post(url)

    def list_smtp(self):
        """
        ``GET /api/v2/admin/smtp-settings``

        `Admin List SMTP Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#list-smtp-settings>`_
        """
        url = f"{self._endpoint_base_url}/smtp-settings"
        return self._list(url)

    def update_smtp(self, data):
        """
        ``PATCH /api/v2/admin/smtp-settings``

        `Admin Update SMTP Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#update-smtp-settings>`_
        """
        url = f"{self._endpoint_base_url}/smtp-settings"
        return self._patch(url, data)

    def list_twilio(self):
        """
        ``GET /api/v2/admin/twilio-settings``

        `Admin List Twilio Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#list-twilio-settings>`_
        """
        url = f"{self._endpoint_base_url}/twilio-settings"
        return self._list(url)

    def update_twilio(self, data):
        """
        ``PATCH /api/v2/admin/twilio-settings``

        `Admin Update Twilio Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#update-twilio-settings>`_
        """
        url = f"{self._endpoint_base_url}/twilio-settings"
        return self._patch(url, data)

    def verify_twilio(self, data):
        """
        ``POST /api/v2/admin/twilio-settings/verify``

        `Admin Verify Twilio Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#verify-twilio-settings>`_
        """
        url = f"{self._endpoint_base_url}/twilio-settings/verify"
        return self._post(url, data=data)

    def list_customization(self):
        """
        ``GET /api/v2/admin/customization-settings``

        `Admin List Customization Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#list-customization-settings>`_
        """
        url = f"{self._endpoint_base_url}/customization-settings"
        return self._list(url)

    def update_customization(self, data):
        """
        ``PATCH /api/v2/admin/customization-settings``

        `Admin Update Customization Settings API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/settings.html#update-customization-settings>`_
        """
        url = f"{self._endpoint_base_url}/customization-settings"
        return self._patch(url, data)



"""
Module for Terraform Cloud API Endpoint: Registry Modules.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCRegistryModules(TFCEndpoint):
    """
    `Registry Modules API Docs (Private Registry) \
        <https://www.terraform.io/docs/cloud/api/modules.html>`_

    `Registry Modules API Docs (Public Registry) \
        <https://www.terraform.io/docs/registry/api.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._modules_v2_base_url = f"{self._api_v2_base_url}/registry-modules"
        self._modules_v1_base_url = f"{self._modules_v1_base_url}"
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations"

    def required_entitlements(self):
        return [Entitlements.PRIVATE_MODULE_REGISTRY]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    # Public Registry API Endpoints
    def list(self, offset=None, limit=None, provider=None, verified=None):
        """
        ``GET <base_url>``
        ``GET <base_url>/:namespace``

        `Registry Modules List API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#list-modules>`_
        """
        # TODO: the slash here is behaving differently from TFE to TFC
        url = f"{self._modules_v1_base_url}/{self._org_name}"
        return self._list(\
            url, offset=offset, limit=limit, provider=provider, verified=verified)

    def search(self, query, offset=None, limit=None, provider=None, verified=None):
        """
        ``GET <base_url>/search``

        `Registry Modules Search API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#search-modules>`_
        """
        # TODO: the slash here is behaving differently from TFE to TFC
        url = f"{self._modules_v1_base_url}/search"
        return self._list(url, \
            query=query, offset=offset, limit=limit, provider=provider,\
            verified=verified)

    def show(self, module_name, provider):
        """
        ``GET /registry-modules/show/:organization_name/:name/:provider``

        `Registry Modules Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#show-a-module>`_
        """
        url = f"{self._modules_v2_base_url}/show/{self._org_name}/{module_name}/{provider}"
        return self._show(url)

    def list_versions(self, name, provider):
        """
        ``GET <base_url>/:namespace/:name/:provider/versions``

        `Registry Modules List Versions API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#list-available-versions-for-a-specific-module>`_
        """
        url = f"{self._modules_v1_base_url}/{self._org_name}/{name}/{provider}/versions"
        return self._get(url)

    def list_latest_version_all_providers(self, name, offset=None, limit=None):
        """
        ``GET <base_url>/:namespace/:name``

        `Registry Modules List Latest Version All Providers API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#list-latest-version-of-module-for-all-providers>`_
        """
        url = f"{self._modules_v1_base_url}/{self._org_name}/{name}"
        return self._list(url, offset=offset, limit=limit)

    def list_latest_version_specific_provider(self, name, provider):
        """
        ``GET <base_url>/:namespace/:name/:provider``

        `Registry Modules List Latest Version Specific Provider API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#latest-version-for-a-specific-module-provider>`_
        """
        url = f"{self._modules_v1_base_url}/{self._org_name}/{name}/{provider}"
        return self._get(url)

    def get(self, name, provider, version):
        """
        ``GET <base_url>/:namespace/:name/:provider/:version``

        `Registry Modules Get API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#get-a-specific-module>`_
        """
        url = f"{self._modules_v1_base_url}/{self._org_name}/{name}/{provider}/{version}"
        return self._get(url)

    def download_version_source(self, name, provider, version, target_path):
        """
        ``GET <base_url>/:namespace/:name/:provider/:version/download``

        `Registry Modules Download Version Source API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#download-source-code-for-a-specific-module-version>`_
        """
        url = f"{self._modules_v1_base_url}/{self._org_name}/{name}/{provider}/{version}/download"
        return self._download(url, target_path, header_with_url="X-Terraform-Get")

    def download_latest_source(self, name, provider, target_path):
        """
        ``GET <base_url>/:namespace/:name/:provider/download``

        `Registry Modules Download Latest Source API Doc Reference \
            <https://www.terraform.io/docs/registry/api.html#download-the-latest-version-of-a-module>`_
        """
        url = f"{self._modules_v1_base_url}/{self._org_name}/{name}/{provider}/download"
        return self._download(\
            url, target_path, header_with_url="X-Terraform-Get", allow_redirects=True)

    # Private Registry API Endpoints
    def publish_from_vcs(self, payload):
        """
        ``POST /registry-modules``

        `Registry Modules Publish From VCS API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#publish-a-module-from-a-vcs>`_

        `Publish From VCS Sample Payload \
            <https://www.terraform.io/docs/cloud/api/modules.html#sample-payload>`_
        """
        return self._post(self._modules_v2_base_url, data=payload)

    def destroy(self, module_name, provider=None, version=None):
        """
        ``POST /registry-modules/actions/delete/:organization_name/:name/:provider/:version``
        ``POST /registry-modules/actions/delete/:organization_name/:name/:provider``
        ``POST /registry-modules/actions/delete/:organization_name/:name``

        `Registry Modules Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#delete-a-module>`_
        """
        url = f"{self._modules_v2_base_url}/actions/delete/{self._org_name}/{module_name}"

        if provider:
            url += f"/{provider}"
            if version:
                url += f"/{version}"

        return self._post(url)

    def create(self, payload):
        """
        ``POST /organizations/:organization_name/registry-modules``

        `Registry Modules Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#create-a-module>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/modules.html#request-body-1>`_
        """

        url = f"{self._org_api_v2_base_url}/{self._org_name}/registry-modules"
        return self._post(url, data=payload)


    def create_version(self, module_name, provider, payload):
        """
        ``POST /registry-modules/:organization_name/:name/:provider/versions``

        `Registry Modules Create Version API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#create-a-module-version>`_

        `Create Version Sample Payload \
            <https://www.terraform.io/docs/cloud/api/modules.html#request-body-2>`_
        """

        url = f"{self._modules_v2_base_url}/{self._org_name}/{module_name}/{provider}/versions"
        return self._post(url, data=payload)

    def upload_version(self, path_to_tarball, upload_url):
        """
        ``PUT https://archivist.terraform.io/v1/object/<UNIQUE OBJECT ID>``

        `Registry Modules Upload Version API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/modules.html#upload-a-module-version>`_
        """
        data = None

        with open(path_to_tarball, 'rb') as data_bytes:
            data = data_bytes.read()

        return self._put(upload_url, data=data)



"""
Module for testing the Terraform Cloud API Endpoint: IP Ranges.
"""

from .base import TestTFCBaseTestCase


class TestTFCIPRanges(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: IP Ranges.
    """

    _unittest_name = "iprang"
    _endpoint_being_tested = "ip_ranges"

    def test_ip_ranges(self):
        """
        Test the IP Ranges API endpoints.
        """
        ip_ranges = self._api.ip_ranges.list()

        self.assertIn("notifications", ip_ranges)
        self.assertIn("sentinel", ip_ranges)
        self.assertIn("vcs", ip_ranges)



"""
Module for Terraform Cloud API Endpoint: Policy Sets.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements, MAX_PAGE_SIZE

class TFCPolicySets(TFCEndpoint):
    """
    `Policy Sets API Docs \
        <https://www.terraform.io/docs/cloud/api/policy-sets.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/policy-sets"
        self._pol_set_version_api_v2_base_url = f"{self._api_v2_base_url}/policy-set-versions"
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations/{org_name}/policy-sets"

    def required_entitlements(self):
        return [Entitlements.SENTINEL]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST /organizations/:organization_name/policy-sets``

        `Policy Sets Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#create-a-policy-set>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#sample-payload>`_
        """
        return self._create(self._org_api_v2_base_url, payload)

    def list(self, filters=None, include=None, page=None, page_size=None, search=None):
        """
        ``GET /organizations/:organization_name/policy-sets``

        `Policy Sets List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#list-policy-sets>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#query-parameters>`_):
            - ``filter[versioned]`` (Optional)
            - ``include`` (Optional)
            - ``page`` (Optional)
            - ``page_size`` (Optional)
            - ``search`` (Optional)

        Example filter(s):

        .. code-block:: python

            filters = [
                {
                    "keys": ["versioned"],
                    "value": "foo"
                }
            ]
        """
        return self._list(\
            self._org_api_v2_base_url, \
            filters=filters, include=include, \
            page=page, page_size=page_size, search=search)

    def list_all(self, search=None, filters=None, include=None):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every policy set for an organization.

        Returns an array of objects.
        """
        current_page_number = 1
        policy_sets_resp = \
            self._list(self._org_api_v2_base_url, \
                page=current_page_number, page_size=MAX_PAGE_SIZE, search=search)
        total_pages = policy_sets_resp["meta"]["pagination"]["total-pages"]

        policy_sets = []
        while current_page_number <= total_pages:
            policy_sets_resp = \
                self._list(\
                    self._org_api_v2_base_url, page=current_page_number, page_size=MAX_PAGE_SIZE, \
                        include=include, filters=filters, search=search)
            policy_sets += policy_sets_resp["data"]
            current_page_number += 1

        return policy_sets

    def show(self, policy_set_id):
        """
        ``GET /policy-sets/:id``

        `Policy Sets Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#show-a-policy-set>`_
        """
        url = f"{self._endpoint_base_url}/{policy_set_id}"
        return self._show(url)

    def update(self, policy_set_id, payload):
        """
        ``PATCH /policy-sets/:id``

        `Policy Sets Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#update-a-policy-set>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{policy_set_id}"
        return self._update(url, payload)

    def destroy(self, policy_set_id):
        """
        ``DELETE /policies/:policy_set_id``

        `Policy Sets Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#delete-a-policy-set>`_
        """
        url = f"{self._endpoint_base_url}/{policy_set_id}"
        return self._destroy(url)

    def add_policies_to_set(self, policy_set_id, payload):
        """
        ``POST /policy-sets/:id/relationships/policies``

        `Policy Sets Add Policies to Set API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#add-policies-to-the-policy-set>`_

        `Add Policies to Set Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#sample-payload-2>`_
        """
        url = f"{self._endpoint_base_url}/{policy_set_id}/relationships/policies"
        return self._post(url, data=payload)

    def attach_policy_set_to_workspaces(self, policy_id, payload):
        """
        ``POST /policy-sets/:id/relationships/workspaces``

        `Policy Sets Attach Set to Workspaces API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#attach-a-policy-set-to-workspaces>`_

        `Attach Policy Set to Workspaces Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#sample-payload-3>`_
        """
        url = f"{self._endpoint_base_url}/{policy_id}/relationships/workspaces"
        return self._post(url, data=payload)

    def remove_policies_from_set(self, policy_id, payload):
        """
        ``DELETE /policy-sets/:id/relationships/policies``

        `Policy Sets Remove Policies From Set API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#remove-policies-from-the-policy-set>`_

        `Remove Policies from Set Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#sample-payload-4>`_
        """
        url = f"{self._endpoint_base_url}/{policy_id}/relationships/policies"
        return self._delete(url, data=payload)

    def detach_policy_set_from_workspaces(self, policy_id, payload):
        """
        ``DELETE /policy-sets/:id/relationships/workspaces``

        `Policy Sets Detach Set from Workspaces API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#detach-the-policy-set-from-workspaces>`_

        `Detach Policy Set From Workspaces Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#sample-payload-5>`_
        """
        url = f"{self._endpoint_base_url}/{policy_id}/relationships/workspaces"
        return self._delete(url, data=payload)

    def create_policy_set_version(self, policy_set_id):
        """
        ``POST /policy-sets/:id/versions``

        `Policy Sets Create Set Version API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#create-a-policy-set-version>`_
        """
        url = f"{self._endpoint_base_url}/{policy_set_id}/versions"
        return self._post(url)

    def show_policy_set_version(self, policy_set_version_id):
        """
        ``GET /policy-set-versions/:id``

        `Policy Sets Show Policy Set Version API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#show-a-policy-set-version>`_
        """
        url = f"{self._pol_set_version_api_v2_base_url}/{policy_set_version_id}"
        return self._get(url)

    def upload(self, path_to_tarball, policy_set_version_id):
        """
        ``PUT {derived_policy_set_upload_url}``
        ``PUT https://archivist.terraform.io/v1/object/<UNIQUE OBJECT ID>``

        `Policy Sets Upload API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-sets.html#upload-policy-set-versions>`_
        """
        url = self.show_policy_set_version(policy_set_version_id)["data"]["links"]["upload"]
        data = None
        with open(path_to_tarball, 'rb') as data_bytes:
            data = data_bytes.read()

        return self._put(url, data=data)



"""
Module for testing the Terraform Cloud API Endpoint: Agents.
"""

from terrasnek.exceptions import TFCHTTPUnprocessableEntity
from .base import TestTFCBaseTestCase


class TestTFCAgents(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Agents.
    """

    _unittest_name = "agents"
    _endpoint_being_tested = "agents"

    def test_agents(self):
        """
        Test the Agent Pools API endpoints.
        """

        # Create an agent pool, we won't assert anything on it, in case it has already been created.
        create_payload = {
            "data": {
                "type": "agent-pools",
                "attributes": {
                    "name": self._random_name()
                }
            }
        }

        try:
            self._api.agents.create_pool(create_payload)
        except TFCHTTPUnprocessableEntity as exc:
            # NOTE: this is not a good test, but the API endpoint isn't very flexible.
            self._logger.info("\
                Unprocessable entity for agent pool. You may already have one created.")
            self._logger.error(exc)

        # List the agent pools, assert that we have only one.
        agent_pools = self._api.agents.list_pools()["data"]
        self.assertEqual(len(agent_pools), 1)

        agent_pool_id = agent_pools[0]["id"]
        shown_agent_pool = self._api.agents.show_pool(agent_pool_id)["data"]
        self.assertEqual(agent_pool_id, shown_agent_pool["id"])

        # Update the agent pool name, check that it has taken effect.
        updated_agent_pool_name = self._random_name()
        update_payload = {
            "data": {
                "type": "agent-pools",
                "attributes": {
                    "name": updated_agent_pool_name
                }
            }
        }
        updated_agent_pool = self._api.agents.update(agent_pool_id, update_payload)
        self.assertEqual(updated_agent_pool_name, updated_agent_pool["data"]["attributes"]["name"])

        # Destroy the agent pool, confirm that it's gone.
        self._api.agents.destroy(agent_pool_id)
        agent_pools = self._api.agents.list_pools()["data"]
        self.assertEqual(len(agent_pools), 0)



"""
Module for Terraform Cloud API Endpoint: Invoices.
"""

from .endpoint import TFCEndpoint

class TFCInvoices(TFCEndpoint):
    """
    `Invoices API Docs \
        <https://www.terraform.io/docs/cloud/api/invoices.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._org_base_url = \
            f"{self._api_v2_base_url}/organizations/{org_name}/invoices"

    def required_entitlements(self):
        # TODO: self serve billing?
        return []

    def terraform_cloud_only(self):
        return True

    def terraform_enterprise_only(self):
        return False

    def list(self):
        """
        ``GET /organizations/:organization_name/invoices``

        `Invoices List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/invoices.html#list-invoices>`_
        """
        # TODO: cursor pagination
        return self._list(self._org_base_url)

    def next(self):
        """
        ``GET /organizations/:organization_name/invoices/next``

        `Invoices Get Next API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/invoices.html#get-next-invoice>`_
        """
        url = f"{self._org_base_url}/next"
        return self._get(url)



"""
Module for testing the Terraform Cloud API Endpoint: Policies.
"""

from .base import TestTFCBaseTestCase


class TestTFCPolicies(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Policies.
    """

    _unittest_name = "pol"
    _endpoint_being_tested = "policies"

    def test_policies(self):
        """
        Test the Policies API endpoints.
        """

        # Confirm we have no policies.
        all_policies = self._api.policies.list()["data"]
        self.assertEqual(len(all_policies), 0)

        # Add a policy to TFC
        created_policy = self._api.policies.create(self._get_policy_create_payload())["data"]
        created_policy_id = created_policy["id"]
        created_policy_name = created_policy["attributes"]["name"]

        # List all the policies, search the policy we just created so
        # we can test out the list params
        some_policies = self._api.policies.list(\
            page=0, page_size=50, search=created_policy_name)["data"]

        found_pol = False
        for pol in some_policies:
            if created_policy_id == pol["id"]:
                found_pol = True
                break
        self.assertTrue(found_pol)

        all_policies = self._api.policies.list_all(search=created_policy_name)

        found_pol = False
        for pol in all_policies:
            if created_policy_id == pol["id"]:
                found_pol = True
                break
        self.assertTrue(found_pol)

        # Upload the policy
        policy_payload = "main = rule { true }"
        self._api.policies.upload(created_policy_id, policy_payload)

        # Show the policy we just created
        shown_policy = self._api.policies.show(created_policy_id)["data"]
        shown_policy_id = shown_policy["id"]

        # Get the text of the policy we just created
        policy_text = self._api.policies.get_policy_text(created_policy_id)
        self.assertEqual(policy_text, policy_payload)

        # Make sure the IDs are the same
        self.assertEqual(shown_policy_id, created_policy_id)

        # And make sure our upload took effect
        self.assertIn("upload", shown_policy["links"])

        # Update the policy
        desc_to_update_to = "foo"
        update_payload = {
            "data": {
                "attributes": {
                    "description": desc_to_update_to
                },
                "type":"policies"
            }
        }
        updated_policy = self._api.policies.update(created_policy_id, update_payload)["data"]
        updated_desc = updated_policy["attributes"]["description"]
        self.assertEqual(desc_to_update_to, updated_desc)

        # Delete the policy we just updated
        self._api.policies.destroy(created_policy_id)

        # Check that we now have zero policies again
        some_policies = self._api.policies.list()["data"]
        self.assertEqual(len(some_policies), 0)



"""
Module for testing the Terraform Cloud API Endpoint: Users.
"""

from .base import TestTFCBaseTestCase


class TestTFCUsers(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Users.
    """

    _unittest_name = "users"
    _endpoint_being_tested = "users"

    def test_users(self):
        """
        Test the Users API endpoints.
        """

        logged_in_user = self._api.account.show()["data"]
        logged_in_user_id = logged_in_user["id"]

        shown_user = self._api.users.show(logged_in_user_id)["data"]
        self.assertEqual(logged_in_user_id, shown_user["id"])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="terrasnek",
    version="0.1.2",
    author="Neil Dahlke",
    author_email="neil.dahlke+terrasnek@gmail.com",
    description="A Python client for the Terraform Cloud API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dahlke/terrasnek",
    packages=setuptools.find_packages(),
    install_requires=[
        "requests>=2.21.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
    ],
)



"""
Constants for default values in the TFC API implementation testing.
"""

import os
import sys
import logging

# Test Defaults
API_LOG_LEVEL = logging.INFO


# Configurable Constants
TFC_URL = os.getenv("TFC_URL", None)
TFC_TOKEN = os.getenv("TFC_TOKEN", None)
TFC_ORG_TOKEN = os.getenv("TFC_ORG_TOKEN", None)

AWS_ACCESS_KEY_ID = os.getenv("AWS_ACCESS_KEY_ID", None)
AWS_SECRET_ACCESS_KEY = os.getenv("AWS_SECRET_ACCESS_KEY", None)

SSL_VERIFY = os.getenv("SSL_VERIFY", "").lower() != "false"

TWILIO_ACCOUNT_SID = os.getenv("TWILIO_ACCOUNT_SID", None)
TWILIO_AUTH_TOKEN = os.getenv("TWILIO_AUTH_TOKEN", None)

SMTP_HOST = os.getenv("SMTP_HOST", None)
SMTP_PORT = os.getenv("SMTP_PORT", None)
SMTP_USERNAME = os.getenv("SMTP_USERNAME", None)
SMTP_PASSWORD = os.getenv("SMTP_PASSWORD", None)

GITHUB_TOKEN = os.getenv("GITHUB_TOKEN", None)
GITHUB_SECRET = os.getenv("GITHUB_SECRET", None)

# NOTE: This is an optional env var, if not provided a random test org will be generated
TEST_ORG_NAME = os.getenv("TEST_ORG_NAME", None)

TEST_EMAIL = os.getenv("TEST_EMAIL", None)
TEST_USERNAME = os.getenv("TEST_USERNAME", None)
TEST_TEAM_NAME = os.getenv("TEST_TEAM_NAME", None)
TEST_PASSWORD = os.getenv("TEST_PASSWORD", None)

if TFC_URL is None:
    sys.exit("Environment variable TFC_URL must be set.")

if TFC_TOKEN is None:
    sys.exit("Environment variable TFC_TOKEN must be set.")

if AWS_ACCESS_KEY_ID is None:
    sys.exit("Environment variable AWS_ACCESS_KEY_ID must be set.")

if AWS_SECRET_ACCESS_KEY is None:
    sys.exit("Environment variable AWS_SECRET_ACCESS_KEY must be set.")

if TWILIO_ACCOUNT_SID is None:
    sys.exit("Environment variable TWILIO_ACCOUNT_SID must be set.")

if TWILIO_AUTH_TOKEN is None:
    sys.exit("Environment variable TWILIO_AUTH_TOKEN must be set.")

if SMTP_HOST is None:
    sys.exit("Environment variable SMTP_HOST must be set.")

if SMTP_PORT is None:
    sys.exit("Environment variable SMTP_PORT must be set.")

if SMTP_USERNAME is None:
    sys.exit("Environment variable SMTP_USERNAME must be set.")

if SMTP_PASSWORD is None:
    sys.exit("Environment variable SMTP_PASSWORD must be set.")

if GITHUB_TOKEN is None:
    sys.exit("Environment variable GITHUB_TOKEN must be set.")

if GITHUB_SECRET is None:
    sys.exit("Environment variable GITHUB_SECRET must be set.")

if TEST_EMAIL is None:
    sys.exit("Environment variable TEST_EMAIL must be set.")

if TEST_USERNAME is None:
    sys.exit("Environment variable TEST_USERNAME must be set.")

if TEST_TEAM_NAME is None:
    sys.exit("Environment variable TEST_TEAM_NAME must be set.")

if TEST_PASSWORD is None:
    sys.exit("Environment variable TEST_PASSWORD must be set.")


# Non-Configurable Constants
TFE_MODULE_PROVIDER_TYPE = "tfe"
DEFAULT_VCS_WORKING_DIR = "tfe"
MAX_TEST_TIMEOUT = 300



"""
Module for testing the Terraform Cloud API Endpoint: Agent Tokens.
"""

from terrasnek.exceptions import TFCHTTPUnprocessableEntity
from .base import TestTFCBaseTestCase


class TestTFCAgentTokens(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Agent Tokens.
    """

    _unittest_name = "agt-tokens"
    _endpoint_being_tested = "agent_tokens"

    def setUp(self):
        # Make sure that we have an Agent Pool on the org if it hasn't already been enabled.
        create_payload = {
            "data": {
                "type": "agent-pools",
                "attributes": {
                    "name": self._random_name()
                }
            }
        }

        try:
            self._api.agents.create_pool(create_payload)
        except TFCHTTPUnprocessableEntity as exc:
            # NOTE: this is not a good test, but the API endpoint isn't very flexible.
            self._logger.info("\
                Unprocessable entity for agent pool. You may already have one created.")
            self._logger.error(exc)

        agent_pools = self._api.agents.list_pools()["data"]
        self._agent_pool_id = agent_pools[0]["id"]

    def test_agent_tokens(self):
        """
        Test the Agent Tokens API endpoints.
        """

        # List the tokens, assert we have zero
        agent_tokens = self._api.agent_tokens.list(self._agent_pool_id)["data"]
        self.assertEqual(len(agent_tokens), 0)

        # Create an agent token, store it's ID
        create_payload = {
            "data": {
                "type": "authentication-tokens",
                "attributes": {
                    "description": "foo"
                }
            }
        }
        created_agent_token = \
            self._api.agent_tokens.create(self._agent_pool_id, create_payload)["data"]
        created_agent_token_id = created_agent_token["id"]

        # List the agent tokens again, assert we have one
        agent_tokens = self._api.agent_tokens.list(self._agent_pool_id)["data"]
        self.assertEqual(len(agent_tokens), 1)

        shown_agent_token = self._api.agent_tokens.show(created_agent_token_id)["data"]
        self.assertEqual(shown_agent_token["id"], created_agent_token_id)

        # Destroy the token, assert we have zero again
        self._api.agent_tokens.destroy(created_agent_token_id)
        agent_tokens = self._api.agent_tokens.list(self._agent_pool_id)["data"]
        self.assertEqual(len(agent_tokens), 0)



"""
Module for Terraform Cloud API Endpoint: Workspace Variables.
"""

from .endpoint import TFCEndpoint

class TFCWorkspaceVars(TFCEndpoint):
    """
    `Workspace Vars API Docs \
        <https://www.terraform.io/docs/cloud/api/workspace-variables.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/workspaces"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, workspace_id, payload):
        """
        ``POST /workspaces/:workspace_id/vars``

        `Workspace Vars Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspace-variables.html#create-a-variable>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/workspace-variables.html#sample-payload>`_
        """
        url = f"{self._endpoint_base_url}/{workspace_id}/vars/"
        return self._create(url, payload)

    def list(self, workspace_id):
        """
        ``GET /workspaces/:workspace_id/vars``

        `Workspace Vars List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspace-variables.html#list-variables>`_
        """
        url = f"{self._endpoint_base_url}/{workspace_id}/vars/"
        return self._list(url)

    def update(self, workspace_id, variable_id, payload):
        """
        ``PATCH /workspaces/:workspace_id/vars/:variable_id``

        `Workspace Vars Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspace-variables.html#update-variables>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/workspace-variables.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{workspace_id}/vars/{variable_id}"
        return self._update(url, payload)

    def destroy(self, workspace_id, variable_id):
        """
        ``DELETE /workspaces/:workspace_id/vars/:variable_id``

        `Workspace Vars Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/workspace-variables.html#delete-variables>`_
        """
        url = f"{self._endpoint_base_url}/{workspace_id}/vars/{variable_id}"
        return self._destroy(url)



"""
Module for Terraform Cloud API Endpoint: Runs.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements, MAX_PAGE_SIZE

class TFCRuns(TFCEndpoint):
    """
    `Runs API Docs \
        <https://www.terraform.io/docs/cloud/api/run.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._ws_api_v2_base_url = f"{self._api_v2_base_url}/workspaces"
        self._runs_api_v2_base_url = f"{self._api_v2_base_url}/runs"

    def required_entitlements(self):
        return [Entitlements.OPERATIONS]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def list(self, workspace_id, page=None, page_size=None):
        """
        ``GET /workspaces/:workspace_id/runs``

        `Runs List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run.html#list-runs-in-a-workspace>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/run.html#query-parameters>`_):
            - ``page`` (Optional)
            - ``page_size`` (Optional)
        """

        url = f"{self._ws_api_v2_base_url}/{workspace_id}/runs"
        return self._list(url, page=page, page_size=page_size)

    def list_all(self, workspace_id):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every run for a workspace.

        Returns an array of objects.
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/runs"

        current_page_number = 1
        runs_resp = self._list(url, page=current_page_number, page_size=MAX_PAGE_SIZE)
        total_pages = runs_resp["meta"]["pagination"]["total-pages"]

        runs = []
        while current_page_number <= total_pages:
            runs_resp = \
                self._list(url, page=current_page_number, page_size=MAX_PAGE_SIZE)
            runs += runs_resp["data"]
            current_page_number += 1

        return runs


    def show(self, run_id):
        """
        ``GET /runs/:run_id``

        `Runs Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run.html#get-run-details>`_
        """

        url = f"{self._runs_api_v2_base_url}/{run_id}"
        return self._show(url)

    def create(self, payload):
        """
        ``POST /runs``

        `Runs Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run.html#create-a-run>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/run.html#sample-payload>`_
        """

        return self._create(self._runs_api_v2_base_url, payload)

    def apply(self, run_id, payload):
        """
        ``POST /runs/:run_id/actions/apply``

        `Runs Apply API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run.html#apply-a-run>`_
        """

        url = f"{self._runs_api_v2_base_url}/{run_id}/actions/apply"
        return self._post(url, data=payload)

    def discard(self, run_id, payload):
        """
        ``POST /runs/:run_id/actions/discard``

        `Runs Discard API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run.html#discard-a-run>`_
        """

        url = f"{self._runs_api_v2_base_url}/{run_id}/actions/discard"
        return self._post(url, data=payload)

    def cancel(self, run_id, payload):
        """
        ``POST /runs/:run_id/actions/cancel``

        `Runs Cancel API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run.html#cancel-a-run>`_
        """

        url = f"{self._runs_api_v2_base_url}/{run_id}/actions/cancel"
        return self._post(url, data=payload)

    def force_cancel(self, run_id, payload):
        """
        ``POST /runs/:run_id/actions/force-cancel``

        `Runs Force Cancel API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run.html#forcefully-cancel-a-run>`_
        """

        url = f"{self._runs_api_v2_base_url}/{run_id}/actions/force-cancel"
        return self._post(url, data=payload)

    def force_execute(self, run_id):
        """
        ``POST /runs/:run_id/actions/force-execute``

        `Runs Force Execute API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run.html#forcefully-execute-a-run>`_
        """

        url = f"{self._runs_api_v2_base_url}/{run_id}/actions/force-execute"
        return self._post(url)



"""
Module for Terraform Cloud API Endpoint: Team Tokens.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCTeamTokens(TFCEndpoint):
    """
    `Team Tokens API Docs \
        <https://www.terraform.io/docs/cloud/api/team-tokens.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/teams"

    def required_entitlements(self):
        return [Entitlements.TEAMS]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, team_id, payload=None):
        """
        ``POST /teams/:team_id/authentication-token``

        `Team Tokens Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/team-tokens.html#generate-a-new-team-token>`_
        """

        url = f"{self._endpoint_base_url}/{team_id}/authentication-token"
        return self._create(url, payload=payload)


    def destroy(self, team_id):
        """
        ``DELETE /teams/:team_id/authentication-token``

        `Team Tokens Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/team-tokens.html#delete-the-team-token>`_
        """
        url = f"{self._endpoint_base_url}/{team_id}/authentication-token"
        return self._destroy(url)



"""
Module for testing the Terraform Cloud API Endpoint: Admin Runs.
"""

import time

from .base import TestTFCBaseTestCase


class TestTFCAdminRuns(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Admin Runs.
    """

    _unittest_name = "adm-run"
    _endpoint_being_tested = "admin_runs"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        # Store the OAuth client ID to remove it at the end.
        oauth_client = self._api.oauth_clients.create(\
            self._get_oauth_client_create_payload())["data"]
        self._oauth_client_id = oauth_client["id"]
        oauth_token_id = oauth_client["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a workspace using that token ID, save the workspace ID
        _ws_payload = self._get_ws_with_vcs_create_payload(oauth_token_id)
        workspace = self._api.workspaces.create(_ws_payload)["data"]
        self._ws_id = workspace["id"]

        # Configure the required variables on the workspace
        variable_payloads = [
            self._get_variable_create_payload(
                "email", self._test_email, self._ws_id),
            self._get_variable_create_payload(
                "org_name", self._test_org_name, self._ws_id),
            self._get_variable_create_payload(
                "TFE_TOKEN", self._test_api_token, self._ws_id, category="env", sensitive=True)
        ]
        for payload in variable_payloads:
            self._api.vars.create(payload)

        # Sleep for 1 second to give the WS time to create
        time.sleep(1)

        # Start the run, store the run ID
        create_run_payload = self._get_run_create_payload(self._ws_id)
        run = self._api.runs.create(create_run_payload)["data"]
        self._run_id = run["id"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_admin_runs(self):
        """
        Test all the Admin Runs API endpoints.
        """

        # List all the runs confirm the one we created in the setup is there
        all_runs = self._api.admin_runs.list(\
            query=self._run_id, filters=[], page=0, page_size=50)["data"]
        found_run = False
        for run in all_runs:
            run_id = run["id"]
            if run_id == self._run_id:
                found_run = True
                break
        self.assertTrue(found_run)

        # Wait a second for to give the run some time to half plan
        self._logger.debug("Sleeping while plan half-executes...")
        time.sleep(1)
        self._logger.debug("Done sleeping.")

        # Force cancel the run and confirm that it has been cancelled
        self._api.admin_runs.force_cancel(self._run_id)
        status_timestamps = \
            self._api.runs.show(self._run_id)["data"]["attributes"]["status-timestamps"]
        while "force-canceled-at" not in status_timestamps:
            # Wait a second while we wait for the force cancel to go through
            time.sleep(1)
            status_timestamps = \
                self._api.runs.show(self._run_id)["data"]["attributes"]["status-timestamps"]
        self.assertIsNotNone(status_timestamps["force-canceled-at"])



"""
Module for Terraform Cloud API Endpoint: Registry Modules.
"""

import time
import os
import timeout_decorator

from terrasnek.exceptions import TFCHTTPNotFound
from .base import TestTFCBaseTestCase
from ._constants import TFE_MODULE_PROVIDER_TYPE, MAX_TEST_TIMEOUT

class TestTFCRegistryModules(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Registry Modules.
    """

    _unittest_name = "reg-mod"
    _endpoint_being_tested = "registry_modules"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        # Store the OAuth client ID to remove it at the end.
        oauth_client = self._api.oauth_clients.create(\
            self._get_oauth_client_create_payload())["data"]
        self._oauth_client_id = oauth_client["id"]
        self._oauth_token_id = \
            oauth_client["relationships"]["oauth-tokens"]["data"][0]["id"]

    def tearDown(self):
        # Delete all the modules before deleting the OAuth client, otherwise
        # you might not be able to delete it after the oauth client is purged.
        self._purge_module_registry()
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_registry_modules(self):
        """
        Test the Registry Modules API endpoints.
        """

        # Publish a module from the VCS provider, using the OAuth token generated
        # in the class setup. Assert that we got the expected response back.
        publish_payload = {
            "data": {
                "attributes": {
                    "vcs-repo": {
                        "identifier": "dahlke/terraform-tfe-terrasnek-unittest-7",
                        "oauth-token-id": self._oauth_token_id,
                        "display_identifier": "dahlke/terraform-tfe-terrasnek-unittest-7"
                    }
                },
                "type":"registry-modules"
            }
        }
        published_module = \
            self._api.registry_modules.publish_from_vcs(publish_payload)["data"]
        published_module_name = published_module["attributes"]["name"]
        self.assertEqual("registry-modules", published_module["type"])

        # Test the listing of the modules, time out if it takes too long.
        # List all the modules for this org, confirm we found the one we
        # published.
        @timeout_decorator.timeout(MAX_TEST_TIMEOUT)
        def found_module_in_listed_modules_timeout(name_to_check):
            found_module = False
            listed_modules_resp = self._api.registry_modules.list()
            listed_modules = listed_modules_resp["modules"]

            while True:
                listed_modules_resp = self._api.registry_modules.list()
                listed_modules = listed_modules_resp["modules"]

                for module in listed_modules:
                    if module["name"] == name_to_check:
                        found_module = True
                        break

                if found_module:
                    break

                time.sleep(1)
            return listed_modules, found_module
        listed_modules, found_module = found_module_in_listed_modules_timeout(published_module_name)

        self.assertTrue(found_module)

        # Search for the module by name, confirm we got it back in the results.
        search_modules_resp = self._api.registry_modules.search(published_module_name)
        search_modules = search_modules_resp["modules"]
        found_module = False
        for module in search_modules:
            if module["name"] == published_module_name:
                found_module = True
                break
        self.assertTrue(found_module)

        # List the module versions, confirm that we got an expected response.
        listed_versions_resp = \
            self._api.registry_modules.list_versions(\
                published_module_name, TFE_MODULE_PROVIDER_TYPE)
        self.assertIn("modules", listed_versions_resp)
        latest_listed_version = listed_modules[-1]["version"]

        # List the latest version for all providers, compare to the
        # published module version
        listed_latest_version_all_providers = \
            self._api.registry_modules.list_versions(\
                published_module_name, TFE_MODULE_PROVIDER_TYPE)
        latest_all_providers = listed_latest_version_all_providers["modules"][0]
        latest_provider_versions = \
            [provider_version["version"] for provider_version in latest_all_providers["versions"]]
        self.assertIn(latest_listed_version, latest_provider_versions)

        # List the latest version for a specific provider, compare to the
        # published module version
        listed_latest_version_specific_provider = \
            self._api.registry_modules.list_versions(\
                published_module_name, TFE_MODULE_PROVIDER_TYPE)
        latest_specific_provider = listed_latest_version_specific_provider["modules"][0]
        latest_specific_versions = \
            [provider_version["version"] for \
                provider_version in latest_specific_provider["versions"]]
        self.assertIn(latest_listed_version, latest_specific_versions)

        # Download the source for a specific version of the module, confirm the file
        # was downloaded to the correct path (and then remove it).
        self._api.registry_modules.download_version_source(\
            published_module_name, TFE_MODULE_PROVIDER_TYPE, latest_listed_version, \
            self._module_version_source_tarball_target_path)
        self.assertTrue(os.path.exists(self._module_version_source_tarball_target_path))
        if os.path.exists(self._module_version_source_tarball_target_path):
            os.remove(self._module_version_source_tarball_target_path)

        # Download the source for the latest version of the module, confirm the file
        # was downloaded to the correct path (and then remove it).
        self._api.registry_modules.download_latest_source(\
            published_module_name, TFE_MODULE_PROVIDER_TYPE, \
            self._module_latest_source_tarball_target_path)
        self.assertTrue(os.path.exists(self._module_latest_source_tarball_target_path))
        if os.path.exists(self._module_latest_source_tarball_target_path):
            os.remove(self._module_latest_source_tarball_target_path)

        # Get the module, compare to the published module name
        gotten_module = \
            self._api.registry_modules.get(\
                published_module_name, TFE_MODULE_PROVIDER_TYPE, latest_listed_version)
        self.assertEqual(published_module_name, gotten_module["name"])

        # List all the provider versions for the module, confirm they are as expected.
        expected_provider = "tfe"
        latest_version_all_providers = \
            self._api.registry_modules.list_latest_version_all_providers(published_module_name)
        tfe_provider_data = latest_version_all_providers["modules"][0]
        self.assertEqual(tfe_provider_data["version"], "0.0.3")
        self.assertEqual(tfe_provider_data["provider"], expected_provider)

        # Confirm the latest version for specific providers endpoint works as expected
        latest_version_tfe_provider = \
            self._api.registry_modules.list_latest_version_specific_provider(\
                published_module_name, expected_provider)
        self.assertEqual(latest_version_tfe_provider["version"], "0.0.3")
        self.assertEqual(latest_version_tfe_provider["provider"], expected_provider)

        shown_module = \
            self._api.registry_modules.show(\
                published_module_name, TFE_MODULE_PROVIDER_TYPE)["data"]
        self.assertEqual(shown_module["attributes"]["name"], published_module_name)

        # Deleted the published module, confirm that it's gone.
        self._api.registry_modules.destroy(\
            published_module_name, provider=TFE_MODULE_PROVIDER_TYPE)
        self.assertRaises(TFCHTTPNotFound, \
            self._api.registry_modules.get, \
                published_module_name, TFE_MODULE_PROVIDER_TYPE, latest_listed_version)


        new_module_name = self._random_name()
        create_payload = {
            "data": {
                "type": "registry-modules",
                "attributes": {
                    "name": new_module_name,
                    "provider": TFE_MODULE_PROVIDER_TYPE
                }
            }
        }
        created_module = self._api.registry_modules.create(create_payload)["data"]
        self.assertEqual(new_module_name, created_module["attributes"]["name"])

        example_version = "0.0.1"
        create_version_payload = {
            "data": {
                "type": "registry-module-versions",
                "attributes": {
                    "version": example_version
                }
            }
        }
        created_version = \
            self._api.registry_modules.create_version(\
                new_module_name, TFE_MODULE_PROVIDER_TYPE, create_version_payload)["data"]
        self.assertEqual("registry-module-versions", created_version["type"])

        created_version_upload_url = created_version["links"]["upload"]

        uploaded_version_resp = \
            self._api.registry_modules.upload_version(\
                self._module_upload_tarball_path, created_version_upload_url)
        self.assertIsNone(uploaded_version_resp)

        self._api.registry_modules.destroy(\
            new_module_name, provider=TFE_MODULE_PROVIDER_TYPE)



"""
Module for Terraform Cloud API Endpoint: Subscriptions.
"""

from .endpoint import TFCEndpoint

class TFCSubscriptions(TFCEndpoint):
    """
    `Subscriptions API Docs \
        <https://www.terraform.io/docs/cloud/api/subscriptions.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._subscriptions_base_url = \
            f"{self._api_v2_base_url}/subscriptions"
        self._org_base_url = \
            f"{self._api_v2_base_url}/organizations/{org_name}/subscription"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return True

    def terraform_enterprise_only(self):
        return False

    def show(self):
        """
        ``GET /organizations/:organization_name/subscription``

        `Subscriptions Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/subscriptions.html#show-subscription-for-organization>`_
        """
        return self._get(self._org_base_url)

    def show_by_id(self, sub_id):
        """
        ``GET /subscriptions/:id``

        `Subscriptions Show by ID API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/subscriptions.html#show-subscription-by-id>`_
        """
        url = f"{self._subscriptions_base_url}/{sub_id}"
        return self._get(url)



"""
Module for testing the Terraform Cloud API Endpoint: OAuth Clients.
"""

from .base import TestTFCBaseTestCase


class TestTFCOAuthClients(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: OAuth Clients.
    """

    _unittest_name = "oa-clients"
    _endpoint_being_tested = "oauth_clients"

    def test_oauth_clients(self):
        """
        Test the OAuth Clients API endpoints.
        """

        # Create a test OAuth client
        oauth_clients = self._api.oauth_clients.list()["data"]
        num_clients_before_add = len(oauth_clients)
        oauth_client = self._api.oauth_clients.create(self._get_oauth_client_create_payload())

        # List the OAuth clients and confirm there is one.
        oauth_clients = self._api.oauth_clients.list()["data"]
        num_clients_after_add = len(oauth_clients)
        self.assertNotEqual(num_clients_after_add, num_clients_before_add)

        # Confirm we can show that OAuth client with it's ID and it matches the created
        oauth_client_id = oauth_clients[0]["id"]
        oauth_client = self._api.oauth_clients.show(oauth_client_id)["data"]
        self.assertEqual(oauth_client["id"], oauth_client_id)

        key_to_update_to = "foo"
        update_payload = {
            "data": {
                "id": oauth_client_id,
                "type": "oauth-clients",
                "attributes": {
                    "key": key_to_update_to
                }
            }
        }
        updated_oauth_client = \
            self._api.oauth_clients.update(oauth_client_id, update_payload)["data"]
        self.assertEqual(updated_oauth_client["attributes"]["key"], key_to_update_to)

        # Destroy the test OAuth client and confirm there are none left
        self._api.oauth_clients.destroy(oauth_client_id)
        oauth_clients = self._api.oauth_clients.list()["data"]
        num_clients_after_delete = len(oauth_clients)
        self.assertNotEqual(num_clients_after_add, num_clients_after_delete)



"""
Module for testing the Terraform Cloud API Endpoint: Admin Workspaces.
"""

from .base import TestTFCBaseTestCase


class TestTFCAdminWorkspaces(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Admin Workspaces.
    """

    _unittest_name = "adm-ws"
    _endpoint_being_tested = "admin_workspaces"

    def setUp(self):
        # Create a sample workspace to manipulate in the test.
        self._ws = self._api.workspaces.create(self._get_ws_without_vcs_create_payload())["data"]
        self._created_ws_id = self._ws["id"]
        self._created_ws_name = self._ws["attributes"]["name"]

    def test_admin_workspaces(self):
        """
        Test the Admin Workspaces API endpoints.
        """

        # List all the workspaces, confirm the one we created in the setup exists
        all_ws = self._api.admin_workspaces.list(\
            search=self._created_ws_name, filters=[], page=0, page_size=50, sort="name")["data"]
        found_ws = False
        for workspace in all_ws:
            ws_id = workspace["id"]
            if ws_id == self._created_ws_id:
                found_ws = True
                break
        self.assertTrue(found_ws)

        # Show the workspace by the ID, confirm we get the expected ID back
        shown_ws = self._api.admin_workspaces.show(self._created_ws_id)["data"]
        self.assertEqual(self._created_ws_id, shown_ws["id"])

        # Destroy the workspace that we created, verify it's gone
        self._api.admin_workspaces.destroy(self._created_ws_id)
        all_ws = self._api.admin_workspaces.list()["data"]
        found_ws = False
        for workspace in all_ws:
            ws_id = workspace["id"]
            if ws_id == self._created_ws_id:
                found_ws = True
                break
        self.assertFalse(found_ws)



"""
Module for Terraform Cloud API Endpoint: SSH Keys.
"""

from .endpoint import TFCEndpoint

class TFCSSHKeys(TFCEndpoint):
    """
    `SSH Keys API Docs \
        <https://www.terraform.io/docs/cloud/api/ssh-keys.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/ssh-keys"
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations/{org_name}/ssh-keys"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST /organizations/:organization_name/ssh-keys``

        `SSH Keys Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/ssh-keys.html#create-an-ssh-key>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/ssh-keys.html#sample-payload>`_
        """
        return self._create(self._org_api_v2_base_url, payload)

    def list(self):
        """
        ``GET /organizations/:organization_name/ssh-keys``

        `SSH Keys List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/ssh-keys.html#list-ssh-keys>`_
        """
        return self._list(self._org_api_v2_base_url)

    def show(self, ssh_key_id):
        """
        ``GET /ssh-keys/:ssh_key_id``

        `SSH Keys Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/ssh-keys.html#get-an-ssh-key>`_
        """
        url = f"{self._endpoint_base_url}/{ssh_key_id}"
        return self._show(url)

    def update(self, ssh_key_id, payload):
        """
        ``PATCH /ssh-keys/:ssh_key_id``

        `SSH Keys Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/ssh-keys.html#update-an-ssh-key>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/ssh-keys.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{ssh_key_id}"
        return self._update(url, payload)

    def destroy(self, ssh_key_id):
        """
        ``DELETE /ssh-keys/:ssh_key_id``

        `SSH Keys Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/ssh-keys.html#delete-an-ssh-key>`_
        """
        url = f"{self._endpoint_base_url}/{ssh_key_id}"
        return self._destroy(url)



"""
Module for testing the Terraform Cloud API Endpoint: SSH Keys.
"""

from .base import TestTFCBaseTestCase


SSH_KEY_NAME_TO_UPDATE_TO = "foo"

class TestTFCSSHKeys(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: SSH Keys.
    """

    _unittest_name = "ssh-keys"
    _endpoint_being_tested = "ssh_keys"

    def test_ssh_keys(self):
        """
        Test the SSH Keys API endpoints.
        """

        all_ssh_keys = self._api.ssh_keys.list()["data"]
        self.assertEqual(len(all_ssh_keys), 0)

        # Add an SSH Key to TFC
        created_key = self._api.ssh_keys.create(self._get_ssh_key_create_payload())["data"]
        created_key_id = created_key["id"]

        # Check that we now have one key
        all_ssh_keys = self._api.ssh_keys.list()["data"]
        self.assertEqual(len(all_ssh_keys), 1)

        # Show the key we just created
        shown_key = self._api.ssh_keys.show(created_key_id)["data"]
        shown_key_id = shown_key["id"]
        self.assertEqual(shown_key_id, created_key_id)

        # Update the key we just showed
        update_payload = {
            "data": {
                "attributes": {
                    "name": SSH_KEY_NAME_TO_UPDATE_TO
                }
            }
        }
        updated_key = self._api.ssh_keys.update(created_key_id, update_payload)["data"]
        updated_key_name = updated_key["attributes"]["name"]
        self.assertEqual(SSH_KEY_NAME_TO_UPDATE_TO, updated_key_name)

        # Delete the key we just updated, confirm it's gone
        self._api.ssh_keys.destroy(created_key_id)
        all_ssh_keys = self._api.ssh_keys.list()["data"]
        self.assertEqual(len(all_ssh_keys), 0)



"""
Base module for all of the Terraform Cloud API modules.

Contains many of the defaults and payload generators.
"""

import logging
import hashlib
import base64
import unittest
import os
import binascii
import time
import timeout_decorator
import terrasnek.exceptions

from terrasnek.api import TFC
from terrasnek._constants import Entitlements

from ._constants import \
    TFC_TOKEN, TFC_ORG_TOKEN, TFC_URL, TEST_EMAIL, \
    TEST_ORG_NAME, TEST_USERNAME, TEST_TEAM_NAME, \
    GITHUB_TOKEN, GITHUB_SECRET, \
    SSL_VERIFY, TEST_PASSWORD, MAX_TEST_TIMEOUT, \
    DEFAULT_VCS_WORKING_DIR, API_LOG_LEVEL

class TestTFCBaseTestCase(unittest.TestCase):
    """
    Base class for providing common test utilities across API endpoints. It
    includes helpers to generate create payloads for common endpoints used in
    many tests.
    """

    _unittest_name = "base"
    _endpoint_being_tested = None

    @classmethod
    def setUpClass(cls):
        cls._logger = logging.getLogger(cls.__class__.__name__)
        cls._logger.setLevel(API_LOG_LEVEL)
        cls._tfc_url = TFC_URL

        cls._test_api_token = TFC_TOKEN
        cls._test_api_org_token = TFC_ORG_TOKEN
        cls._api_log_level = API_LOG_LEVEL
        cls._ssl_verify = SSL_VERIFY

        cls._api = TFC(\
            cls._test_api_token, url=cls._tfc_url, \
                verify=cls._ssl_verify, log_level=cls._api_log_level)

        cls._test_username = TEST_USERNAME
        cls._test_email = TEST_EMAIL
        cls._test_team_name = TEST_TEAM_NAME
        cls._test_password = TEST_PASSWORD

        cls._test_state_path = "./test/testdata/terraform/terrasnek_unittest.tfstate"
        cls._config_version_upload_tarball_path = \
            "./test/testdata/terraform/terrasnek_unittest_config_version.tar.gz"
        with open(\
            './test/testdata/terraform/terrasnek_unittest_config_version_from_string.tf', 'r') \
                as tf_file:
            cls._config_version_upload_string = tf_file.read()
        cls._module_upload_tarball_path = \
            "./test/testdata/terraform/terrasnek_unittest_module.tar.gz"
        cls._policy_set_upload_tarball_path = \
            "./test/testdata/sentinel/terrasnek_unittest_sentinel.tar.gz"
        cls._plan_export_tarball_target_path = \
            "/tmp/terrasnek_unittest_plan_export.tar.gz"
        cls._plan_json_tarball_target_path = \
            "/tmp/terrasnek_unittest_plan_json.tar.gz"
        cls._module_version_source_tarball_target_path = \
            "/tmp/terrasnek_unittest_module_version_export.tar.gz"
        cls._module_latest_source_tarball_target_path = \
            "/tmp/terrasnek_unittest_module_latest_export.tar.gz"

        # If a test org is specified, use the specified org, otherwise create
        # a new one to run the testing in.
        if TEST_ORG_NAME:
            cls._test_org_name = TEST_ORG_NAME
            cls._test_org = cls._api.orgs.show(cls._test_org_name)
        else:
            cls._test_org_name = cls._random_name()
            org_create_payload = {
                "data": {
                    "type": "organizations",
                    "attributes": {
                        "name": cls._test_org_name,
                        "email": cls._test_email
                    }
                }
            }
            cls._test_org = cls._api.orgs.create(org_create_payload)

        cls._api.set_org(cls._test_org_name)

        # Check to see if this test can be run with the current entitlments
        missing_entitlements = cls._get_missing_entitlements(cls._endpoint_being_tested)

        if missing_entitlements:
            raise unittest.SkipTest(\
                "Missing required Terraform Cloud Entitlments for test", \
                    cls._unittest_name, missing_entitlements)

        endpoint_to_test = getattr(cls._api, cls._endpoint_being_tested)

        if cls._api.is_terraform_cloud() and endpoint_to_test.terraform_enterprise_only():
            raise unittest.SkipTest(\
                (f"Skipping Test (%s), since we're testing against Terraform Cloud." % \
                    cls._endpoint_being_tested))

        if not cls._api.is_terraform_cloud() and endpoint_to_test.terraform_cloud_only():
            raise unittest.SkipTest(\
                (f"Skipping Test (%s), since we're testing against Terraform Enterprise." % \
                    cls._endpoint_being_tested))

        cls._purge_organization()

    @classmethod
    def tearDownClass(cls):
        # Only destroy the org if we auto generated it
        if not TEST_ORG_NAME:
            cls._logger.debug("Destroying autogenerated org %s...", cls._test_org_name)
            cls._api.orgs.destroy(cls._test_org_name)
            cls._logger.debug("Destroyed autogenerated org %s.", cls._test_org_name)

    @classmethod
    def _purge_organization(cls):
        cls._logger.info(\
            f"Purging test org ({cls._test_org_name}) of all resources to start fresh...")

        cls._logger.info(f"Purging test org ({cls._test_org_name}) of workspaces...")
        workspaces = cls._api.workspaces.list_all()
        for workspace in workspaces:
            cls._api.workspaces.destroy(workspace_id=workspace["id"])
        cls._logger.debug(f"Workspaces purged from test org ({cls._test_org_name}).")

        cls._purge_module_registry()

        cls._logger.debug(f"Modules purged from test org ({cls._test_org_name}).")

        cls._logger.debug(f"Purging test org ({cls._test_org_name}) of policies...")
        policies = cls._api.policies.list_all()
        for policy in policies:
            cls._api.policies.destroy(policy["id"])
        cls._logger.debug(f"Policies purged from test org ({cls._test_org_name}).")

        cls._logger.debug(f"Purging test org ({cls._test_org_name}) of policy sets...")
        policy_sets = cls._api.policy_sets.list_all()
        for policy_set in policy_sets:
            cls._api.policy_sets.destroy(policy_set["id"])
        cls._logger.debug(f"Policy sets purged from test org ({cls._test_org_name}).")

        # Delete all the VCS adjacent resources before the VCS client
        cls._logger.debug(f"Purging test org ({cls._test_org_name}) of OAuth clients...")
        oauth_clients = cls._api.oauth_clients.list()["data"]
        for oauth_client in oauth_clients:
            cls._api.oauth_clients.destroy(oauth_client["id"])
        cls._logger.debug(f"OAuth clients purged from test org ({cls._test_org_name}).")

        cls._logger.debug(f"Purging test org ({cls._test_org_name}) of SSH Keys...")
        ssh_keys = cls._api.ssh_keys.list()["data"]
        for ssh_key in ssh_keys:
            cls._api.ssh_keys.destroy(ssh_key["id"])
        cls._logger.debug(f"SSH keys purged from test org ({cls._test_org_name}).")

        # Deleting the teams will delete all team memberships, and team tokens
        cls._logger.debug(f"Purging test org ({cls._test_org_name}) of teams...")
        teams = cls._api.teams.list()["data"]
        for team in teams:
            team_name = team["attributes"]["name"]
            if team_name != "owners":
                cls._api.teams.destroy(team["id"])
        cls._logger.debug(f"Teams purged from test org ({cls._test_org_name}).")

        cls._logger.debug(f"Purging test org ({cls._test_org_name}) of org membership invites...")
        org_memberships = cls._api.org_memberships.list_all_for_org()
        for org_memberships in org_memberships:
            membership_id = org_memberships["id"]
            member_status = org_memberships["attributes"]["status"]
            if member_status == "invited":
                cls._api.org_memberships.remove(membership_id)
        cls._logger.debug(f"Org member invites purged from test org ({cls._test_org_name}).")

        if cls._api.is_terraform_cloud():
            cls._logger.debug(f"Purging test org ({cls._test_org_name}) of agent pools...")
            agent_pools = cls._api.agents.list_pools()["data"]
            for agent_pool in agent_pools:
                cls._api.agents.destroy(agent_pool["id"])
            cls._logger.debug(f"Agent pools purged from test org ({cls._test_org_name}).")

        try:
            cls._logger.debug(f"Purging org token from test org ({cls._test_org_name})...")
            cls._api.org_tokens.destroy()
            cls._logger.debug(f"Org token purged from test org ({cls._test_org_name}).")
        except terrasnek.exceptions.TFCHTTPUnclassified:
            cls._logger.debug(f"No org token exists for test org ({cls._test_org_name})...")

    @classmethod
    def _purge_module_registry(cls):
        cls._logger.debug(f"Purging test org ({cls._test_org_name}) of modules...")
        try:
            registry_modules_resp = cls._api.registry_modules.list()
            registry_modules = registry_modules_resp["modules"]
            for registry_module in registry_modules:
                cls._api.registry_modules.destroy(registry_module["name"])
        except terrasnek.exceptions.TFCHTTPNotFound:
            cls._logger.debug("No modules exist in this org, skipping.")

    @classmethod
    def _get_missing_entitlements(cls, endpoint_attr_name):
        endpoint = getattr(cls._api, endpoint_attr_name)
        required_entitlements = endpoint.required_entitlements()
        current_entitlements = cls._api.get_entitlements()
        missing_entitlements = []

        if cls._api.is_terraform_cloud() and current_entitlements is not None:
            for req_ent in required_entitlements:
                meets_sub_requirement = False

                for cur_ent_key in current_entitlements:
                    ent_enabled = current_entitlements[cur_ent_key]
                    cur_ent_key = cur_ent_key.replace("-", "_").upper()

                    if Entitlements[cur_ent_key] == req_ent and ent_enabled:
                        meets_sub_requirement = True

                if not meets_sub_requirement:
                    missing_entitlements.append(req_ent)

        return missing_entitlements

    @classmethod
    def _random_name(cls, ran_str_len=8):
        random_hex = binascii.b2a_hex(os.urandom(ran_str_len)).decode("ascii")
        return f"terrasnek-unittest-{random_hex}"

    @classmethod
    def _unittest_random_name(cls, ran_str_len=8):
        random_hex = binascii.b2a_hex(os.urandom(ran_str_len)).decode("ascii")
        return f"terrasnek-test-{cls._unittest_name}-{random_hex}"

    @staticmethod
    def _get_config_version_create_payload():
        return {
            "data": {
                "type": "configuration-versions",
                "attributes": {
                    "auto-queue-runs": False
                }
            }
        }

    @staticmethod
    def _get_user_token_create_payload():
        return {
            "data": {
                "type": "authentication-tokens",
                "attributes": {
                    "description": "api"
                }
            }
        }

    @staticmethod
    def _get_run_create_payload(workspace_id):
        return {
            "data": {
                "attributes": {
                    "is-destroy": False,
                    "message": "test"
                },
                "type": "runs",
                "relationships": {
                    "workspace": {
                        "data": {
                            "type": "workspaces",
                            "id": workspace_id
                        }
                    }
                }
            }
        }

    @staticmethod
    def _get_variable_create_payload(\
        key, value, workspace_id, category="terraform", sensitive=False):
        return {
            "data": {
                "type": "vars",
                "attributes": {
                    "key": key,
                    "value": value,
                    "category": category,
                    "hcl": False,
                    "sensitive": sensitive
                },
                "relationships": {
                    "workspace": {
                        "data": {
                            "id": workspace_id,
                            "type": "workspaces"
                        }
                    }
                }
            }
        }

    def _get_state_version_create_payload(self):
        # Go Example:
        # https://github.com/hashicorp/go-tfe/blob/4ca75c88c51753c622df5bf4446e69eff6c885d6/state_version_test.go#L105
        raw_state_bytes = None

        with open(self._test_state_path, "rb") as infile:
            raw_state_bytes = infile.read()

        state_hash = hashlib.md5()
        state_hash.update(raw_state_bytes)
        state_md5 = state_hash.hexdigest()

        state_b64 = base64.b64encode(raw_state_bytes).decode("utf-8")

        return {
            "data": {
                "type": "state-versions",
                "attributes": {
                    "serial": 1,
                    "md5": state_md5,
                    "state": state_b64
                }
            }
        }

    def _get_ws_with_vcs_create_payload(self, oauth_token_id, working_dir=DEFAULT_VCS_WORKING_DIR):
        # NB: Needs to be TF > v0.12 for Cost Estimation to work
        return {
            "data": {
                "attributes": {
                    "name": self._unittest_random_name(),
                    "terraform_version": "0.12.24",
                    "working-directory": working_dir,
                    "vcs-repo": {
                        "identifier": "dahlke/terrasnek-unittest-config",
                        "oauth-token-id": oauth_token_id,
                        "branch": "master"
                    }
                },
                "type": "workspaces"
            }
        }

    def _get_ws_without_vcs_create_payload(self):
        return {
            "data": {
                "type": "workspaces",
                "attributes": {
                    "name": self._unittest_random_name()
                }
            }
        }

    def _get_ssh_key_create_payload(self):
        return {
            "data": {
                "type": "ssh-keys",
                "attributes": {
                    "name": self._unittest_random_name(),
                    "value": "-----BEGIN RSA PRIVATE KEY-----\nfoo..."
                }
            }
        }

    def _get_policy_create_payload(self):
        # https://www.terraform.io/docs/cloud/api/policies.html#sample-payload
        return {
            "data": {
                "attributes": {
                    "enforce": [
                        {
                            "path": "terransek-example-policy.sentinel",
                            "mode": "soft-mandatory"
                        }
                    ],
                    "name": self._unittest_random_name(),
                    "description": "terrasnek example policy"
                },
                "relationships": {
                    "policy-sets": {
                        "data": []
                    }
                },
                "type": "policies"
            }
        }

    def _get_policy_set_create_payload(self, oauth_token_id):
        # https://www.terraform.io/docs/cloud/api/policies.html#sample-payload
        return {
            "data": {
                "type": "policy-sets",
                "attributes": {
                    "name": self._unittest_random_name(),
                    "description": "terrasnek unittest",
                    "global": False,
                    "policies-path": "sentinel/",
                    "vcs-repo": {
                        "branch": "master",
                        "identifier": "dahlke/terrasnek-unittest-config",
                        "ingress-submodules": False,
                        "oauth-token-id": oauth_token_id
                    }
                },
                "relationships": {
                    "policies": {
                        "data": []
                    },
                    "workspaces": {
                        "data": []
                    }
                }
            }
        }

    def _get_org_create_payload(self):
        # https://www.terraform.io/docs/cloud/api/organizations.html#sample-payload
        return {
            "data": {
                "type": "organizations",
                "attributes": {
                    "name": self._unittest_random_name(),
                    "email": self._test_email
                }
            }
        }

    def _get_org_membership_invite_payload(self):
        return {
            "data": {
                "attributes": {
                    "email": self._test_email,
                },
                "relationships": {
                    "teams": {
                        "data": []
                    },
                },
                "type": "organization-memberships"
            }
        }

    def _get_team_create_payload(self):
        return {
            "data": {
                "type": "organizations",
                "attributes": {
                    "name": self._unittest_random_name(),
                    "organization-access": {
                        "manage-workspaces": True,
                        "manage-policies": True,
                        "manage-vcs-settings": True
                    }
                }
            }
        }

    def _get_oauth_client_create_payload(self):
        return {
            "data": {
                "type": "oauth-clients",
                "attributes": {
                    "name": self._unittest_random_name(),
                    "service-provider": "github",
                    "http-url": "https://github.com",
                    "api-url": "https://api.github.com",
                    "secret": GITHUB_SECRET,
                    "oauth-token-string": GITHUB_TOKEN
                }
            }
        }

    @timeout_decorator.timeout(MAX_TEST_TIMEOUT)
    def _created_run_timeout(self, run_id):
        """
        While running the tests, it's possible that a run gets queued, which
        would cause the test suite to stall indefinitely until the queue is cleared.
        This function is meant to keep that to a minimum, and just fail the test
        if we are waiting too long.
        """
        created_run = self._api.runs.show(run_id)["data"]
        while not created_run["attributes"]["actions"]["is-confirmable"]:
            created_run = self._api.runs.show(run_id)["data"]
            self._logger.debug("Waiting for created run to finish planning...")
            time.sleep(1)
        self._logger.debug("Plan successful.")
        return created_run



"""
Module for testing the Terraform Cloud API Endpoint: Team Access.
"""

from .base import TestTFCBaseTestCase


class TestTFCTeamAccess(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Team Access.
    """

    _unittest_name = "team-acc"
    _endpoint_being_tested = "team_access"

    def setUp(self):
        # Create a test team
        self._team = self._api.teams.create(
            self._get_team_create_payload())["data"]
        self._team_id = self._team["id"]

        # Invite a test user to this org, will be removed after
        invite = self._api.org_memberships.invite(self._get_org_membership_invite_payload())
        self._org_membership_id = invite["data"]["id"]

        # Create a test workspace
        workspace = self._api.workspaces.create(self._get_ws_without_vcs_create_payload())["data"]
        self._ws_id = workspace["id"]
        self._ws_name = workspace["attributes"]["name"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_name=self._ws_name)
        self._api.teams.destroy(self._team_id)
        self._api.org_memberships.remove(self._org_membership_id)

    def test_team_access(self):
        """
        Test the Team Access API endpoints.
        """

        # Create new Team access, confirm it has been created
        team_access_create_payload = {
            "data": {
                "type": "team-workspaces",
                "attributes": {
                    "access": "admin"
                },
                "relationships": {
                    "workspace": {
                        "data": {
                            "type": "workspaces",
                            "id": self._ws_id
                        }
                    },
                    "team": {
                        "data": {
                            "type": "teams",
                            "id": self._team_id
                        }
                    }
                }
            }
        }
        access = self._api.team_access.add_team_access(
            team_access_create_payload)
        access_id = access["data"]["id"]

        # Show the newly created team access, confirm the ID matches to the created one
        shown_access = self._api.team_access.show(access_id)
        self.assertEqual(shown_access["data"]["id"], access_id)

        # Update team access
        update_payload = {
            "data": {
                "attributes": {
                    "access": "custom",
                    "state-versions": "none"
                }
            }
        }
        updated_access = self._api.team_access.update(access_id, update_payload)["data"]
        self.assertEqual(updated_access["attributes"]["state-versions"], "none")

        list_filters = [
            {
                "keys": ["workspace", "id"],
                "value": self._ws_id
            }
        ]
        listed_access = self._api.team_access.list(filters=list_filters)["data"]
        self.assertEqual(updated_access["id"], listed_access[0]["id"])

        # Remove the team access, confirm it's gone
        self._api.team_access.remove_team_access(access_id)
        all_team_access = self._api.workspaces.list()["data"]
        found_team_access = False
        for team_access in all_team_access:
            if team_access["id"] == access_id:
                found_team_access = True
                break
        self.assertFalse(found_team_access)



"""
Module containing class for common endpoint implementations across all TFC Endpoints.
"""

from abc import ABC, abstractmethod

import json
import logging
import requests

from .exceptions import \
    TFCHTTPBadRequest, TFCHTTPUnauthorized, TFCHTTPForbidden, TFCHTTPNotFound, \
        TFCHTTPConflict, TFCHTTPPreconditionFailed, TFCHTTPUnprocessableEntity, \
            TFCHTTPInternalServerError, TFCHTTPUnclassified, TFCHTTPAPIRequestRateLimit

from ._constants import \
    HTTP_OK, HTTP_CREATED, HTTP_ACCEPTED, HTTP_NO_CONTENT, HTTP_BAD_REQUEST, HTTP_UNAUTHORIZED, \
        HTTP_FORBIDDEN, HTTP_NOT_FOUND, HTTP_CONFLICT, HTTP_PRECONDITION_FAILED, \
            HTTP_UNPROCESSABLE_ENTITY, HTTP_API_REQUEST_RATE_LIMIT_REACHED, \
                HTTP_INTERNAL_SERVER_ERROR


class TFCEndpoint(ABC):
    """
    Base class providing common CRUD operation implementations across all TFC Endpoints.
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(log_level)

        # Remove the slack at the end if someone adds it.
        self._instance_url = \
            instance_url if instance_url[-1] != "/" else instance_url[:-1]
        self._api_v2_base_url = f"{self._instance_url}{well_known_paths['tfe.v2'][:-1]}"
        self._meta_base_url = f"{self._instance_url}/api/meta"
        self._modules_v1_base_url = f"{self._instance_url}{well_known_paths['modules.v1'][:-1]}"
        self._headers = headers
        self._org_name = org_name
        self._verify = verify

    @abstractmethod
    def required_entitlements(self):
        """
        Terraform Cloud Entitlements required for endpoint to work.
        """
        return []

    @abstractmethod
    def terraform_cloud_only(self):
        """
        Return ``True`` if this endpoint is only for Terraform Cloud, else ``False``.
        """
        return False

    @abstractmethod
    def terraform_enterprise_only(self):
        """
        Return ``True`` if this endpoint is only for Terraform Enterprise, else ``False``.
        """
        return False

    def _delete(self, url, data=None):
        results = None
        req = requests.delete(\
            url, data=json.dumps(data), headers=self._headers, verify=self._verify)

        if req.status_code == HTTP_NO_CONTENT:
            pass
        elif req.status_code == HTTP_NOT_FOUND:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPNotFound(err)
        elif req.status_code == HTTP_FORBIDDEN:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPForbidden(err)
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        else:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _get(self, url, return_raw=False, allow_redirects=False):
        results = None

        self._logger.debug(f"Trying HTTP GET to URL: {url} ...")
        req = requests.get(\
            url, headers=self._headers, verify=self._verify, allow_redirects=allow_redirects)

        if req.status_code == HTTP_OK and not return_raw:
            results = json.loads(req.content)
            self._logger.debug(f"GET to {url} successful")
        elif req.status_code == HTTP_OK and return_raw:
            results = req.content
        elif req.status_code == HTTP_NO_CONTENT:
            results = req.headers
        elif req.history:
            # NOTE: If we got redirected, run the get on the new URL, and fix the
            # URL to match the private module registry URL schema.
            url = req.url.replace("/v1/modules/", "/api/registry/v1/modules/")
            results = {"redirect-url": url}
        elif req.status_code == HTTP_UNAUTHORIZED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnauthorized(err)
        elif req.status_code == HTTP_NOT_FOUND:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPNotFound(err)
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        else:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _patch(self, url, data=None):
        results = None

        self._logger.debug(f"Trying HTTP PATCH to URL: {url} ...")
        req = requests.patch(url, data=json.dumps(data), headers=self._headers, verify=self._verify)

        if req.status_code == HTTP_OK:
            results = json.loads(req.content)
        elif req.status_code == HTTP_BAD_REQUEST:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPBadRequest(err)
        elif req.status_code == HTTP_UNAUTHORIZED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnauthorized(err)
        elif req.status_code == HTTP_UNPROCESSABLE_ENTITY:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnprocessableEntity(err)
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        else:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _post(self, url, data=None):
        results = None

        self._logger.debug(f"Trying HTTP POST to URL: {url} ...")
        req = requests.post(url, data=json.dumps(data), headers=self._headers, verify=self._verify)

        if req.status_code in [HTTP_OK, HTTP_CREATED]:
            results = json.loads(req.content)
            self._logger.debug(f"POST to {url} successful")
        elif req.status_code in [HTTP_ACCEPTED, HTTP_NO_CONTENT]:
            self._logger.debug(f"POST to {url} successful")
        elif req.status_code == HTTP_BAD_REQUEST:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPBadRequest(err)
        elif req.status_code == HTTP_NOT_FOUND:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPNotFound(err)
        elif req.status_code == HTTP_CONFLICT:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPConflict(err)
        elif req.status_code == HTTP_PRECONDITION_FAILED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPPreconditionFailed(err)
        elif req.status_code == HTTP_UNPROCESSABLE_ENTITY:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnprocessableEntity(err)
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        elif req.status_code == HTTP_INTERNAL_SERVER_ERROR:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPInternalServerError(err)
        else:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _put(self, url, octet=False, data=None):
        results = None

        headers = dict.copy(self._headers)
        if octet is True:
            headers["Content-Type"] = "application/octet-stream"
            data = bytes(data, "utf-8")

        self._logger.debug(f"Trying HTTP PUT to URL: {url} ...")
        req = requests.put(url, data=data, headers=headers, verify=self._verify)

        if req.status_code == HTTP_OK:
            if octet:
                results = json.loads(req.content)
            self._logger.debug(f"PUT to {url} successful")
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        else:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

        return results

    def _create(self, url, payload=None):
        """
        Implementation the common create resource pattern for the TFC API.
        """
        if payload is None:
            payload = {}

        return self._post(url, data=payload)

    def _destroy(self, url):
        """
        Implementation of the common destroy resource pattern for the TFC API.
        """
        req = requests.delete(url, headers=self._headers, verify=self._verify)

        valid_status_codes = [HTTP_OK, HTTP_NO_CONTENT]
        if req.status_code in valid_status_codes:
            self._logger.debug(f"Terraform Cloud resource at URL [{url}] destroyed.")
        elif req.status_code == HTTP_API_REQUEST_RATE_LIMIT_REACHED:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPAPIRequestRateLimit(err)
        else:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)
            raise TFCHTTPUnclassified(err)

    def _list(self, url, query=None, filters=None, \
        page=None, page_size=None, search=None, include=None, sort=None, \
        offset=None, limit=None, provider=None, namespace=None, verified=None, \
        since=None):
        """
        Implementation of the common list resources pattern for the TFC API.
        """

        q_options = []

        if query is not None:
            q_options.append(f"q={query}")

        if filters is not None:
            for fil in filters:
                filter_string = "filter"
                for k in fil["keys"]:
                    filter_string += f"[{k}]"
                filter_string += f"={fil['value']}"
                q_options.append(filter_string)

        if page is not None:
            q_options.append(f"page[number]={page}")

        if page_size is not None:
            q_options.append(f"page[size]={page_size}")

        if search is not None:
            q_options.append(f"page[name]={page_size}")

        if include is not None:
            q_options.append(f"include={include}")

        if sort is not None:
            q_options.append(f"sort={sort}")

        if search is not None:
            q_options.append(f"search[name]={search}")

        if since is not None:
            q_options.append(f"since={since}")

        # V1 Modules API options
        if offset is not None:
            q_options.append(f"offset={offset}")

        if limit is not None:
            q_options.append(f"limit={limit}")

        if provider is not None:
            q_options.append(f"provider={provider}")

        if namespace is not None:
            q_options.append(f"namespace={namespace}")

        if verified is not None:
            q_options.append(f"verified={verified}")

        if q_options:
            url += "?" + "&".join(q_options)

        return self._get(url)

    def _show(self, url):
        """
        Implementation of the common show resource pattern for the TFC API.
        """
        return self._get(url)

    def _update(self, url, payload):
        """
        Implementation of the common update resource pattern for the TFC API.
        """
        return self._patch(url, data=payload)

    def _download(self, url, target_path, header_with_url=None, allow_redirects=False):
        """
        Implementation of a common download pattern from the TFC API.
        """
        results = None
        if header_with_url is None:
            results = self._get(url, return_raw=True, allow_redirects=allow_redirects)
        else:
            response = self._get(url, allow_redirects=allow_redirects)
            if "redirect-url" in response:
                response = self._get(response["redirect-url"])
            download_url = response[header_with_url]
            results = self._get(download_url, return_raw=True)

        with open(target_path, 'wb') as target_file:
            target_file.write(results)

    def get_current_org(self):
        """
        Return the active org for this endpoint.
        """
        return self._org_name



"""
Module for Terraform Cloud API Endpoint: Admin Workspaces.
"""

from .endpoint import TFCEndpoint

class TFCAdminWorkspaces(TFCEndpoint):
    """
    `Admin Workspaces API Docs \
        <https://www.terraform.io/docs/cloud/api/admin/workspaces.html>`_

    The Admin API is exclusive to Terraform Enterprise, and can only be used
    by the admins and operators who install and maintain their organization's
    Terraform Enterprise instance.

    The Workspaces Admin API contains endpoints to help site administrators manage
    workspaces.
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/admin/workspaces"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return True

    def list(self, filters=None, page=None, page_size=None, sort=None, search=None):
        """
        ``GET /admin/workspaces``

        `Admin Workspaces List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/workspaces.html#list-all-workspaces>`_
        """
        return self._list(\
            self._endpoint_base_url, filters=filters, \
            page=page, page_size=page_size, search=search, sort=sort)

    def show(self, ws_id):
        """
        ``GET /admin/workspaces/:id``

        `Admin Workspaces Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/workspaces.html#show-a-workspace>`_
        """
        url = f"{self._endpoint_base_url}/{ws_id}"
        return self._show(url)

    def destroy(self, ws_id):
        """
        ``DELETE /admin/workspaces/:id``

        `Admin Workspaces Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/workspaces.html#destroy-a-workspace>`_
        """
        url = f"{self._endpoint_base_url}/{ws_id}"
        return self._destroy(url)



"""
Module for testing the Terraform Cloud API Endpoint: Subscriptions.
"""

from .base import TestTFCBaseTestCase


class TestTFCSubscriptions(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Subscriptions.
    """

    _unittest_name = "subsc"
    _endpoint_being_tested = "subscriptions"

    def test_subscriptions(self):
        """
        Test the Subscriptions API endpoints.
        """
        subs = self._api.subscriptions.show()["data"]
        sub_id = subs["id"]
        self.assertEqual(subs["type"], "subscriptions")

        sub = self._api.subscriptions.show_by_id(sub_id)["data"]
        self.assertEqual(sub["id"], sub_id)



"""
Module for Terraform Cloud API Endpoint: User Tokens.
"""

from .endpoint import TFCEndpoint

class TFCUserTokens(TFCEndpoint):
    """
    `User Tokens API Docs \
        <https://www.terraform.io/docs/cloud/api/user-tokens.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._users_api_v2_base_url = f"{self._api_v2_base_url}/users"
        self._tokens_api_v2_base_url = f"{self._api_v2_base_url}/authentication-tokens"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, user_id, payload):
        """
        ``POST /users/:user_id/authentication-tokens``

        `User Tokens Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/user-tokens.html#create-a-user-token>`_

        NOTE: the following method descriptor is here for the ``api_comparison.py`` script to pass,
        since it uses a different prefix than the majority of endpoints.

        ``POST /api/v2/users/:user_id/authentication-tokens``

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/user-tokens.html#sample-payload>`_
        """
        url = f"{self._users_api_v2_base_url}/{user_id}/authentication-tokens"
        return self._create(url, payload)

    def destroy(self, token_id):
        """
        ``DELETE /authentication-tokens/:id``

        `User Tokens Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/user-tokens.html#destroy-a-user-token>`_

        NOTE: the following method descriptor is here for the ``api_comparison.py`` script to pass,
        since it uses a different prefix than the majority of endpoints.

        ``DELETE /api/v2/authentication-tokens/:id``
        """
        url = f"{self._tokens_api_v2_base_url}/{token_id}"
        self._destroy(url)

    def list(self, user_id):
        """
        ``GET /users/:user_id/authentication-tokens``

        `User Tokens List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/user-tokens.html#list-user-tokens>`_

        NOTE: the following method descriptor is here for the ``api_comparison.py`` script to pass,
        since it uses a different prefix than the majority of endpoints.

        ``GET /api/v2/users/:user_id/authentication-tokens``
        """
        url = f"{self._users_api_v2_base_url}/{user_id}/authentication-tokens"
        return self._list(url)

    def show(self, token_id):
        """
        ``GET /authentication-tokens/:id``

        `User Tokens Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/user-tokens.html#show-a-user-token>`_

        NOTE: the following method descriptor is here for the ``api_comparison.py`` script to pass,
        since it uses a different prefix than the majority of endpoints.

        ``GET /api/v2/authentication-tokens/:id``
        """
        url = f"{self._tokens_api_v2_base_url}/{token_id}"
        return self._show(url)



"""
Module for Terraform Cloud API Endpoint: OAuth Clients.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCOAuthClients(TFCEndpoint):
    """
    `OAuth Clients API Docs \
        <https://www.terraform.io/docs/cloud/api/oauth-clients.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._org_api_v2_base_url = \
            f"{self._api_v2_base_url}/organizations/{org_name}/oauth-clients"
        self._oauth_clients_api_v2_base_url = f"{self._api_v2_base_url}/oauth-clients"

    def required_entitlements(self):
        return [Entitlements.VCS_INTEGRATIONS]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def list(self):
        """
        ``GET /organizations/:organization_name/oauth-clients``

        `OAuth Clients List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/oauth-clients.html#list-oauth-clients>`_
        """
        return self._list(self._org_api_v2_base_url)

    def show(self, client_id):
        """
        ``GET /oauth-clients/:id``

        `OAuth Clients Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/oauth-clients.html#show-an-oauth-client>`_
       """
        url = f"{self._oauth_clients_api_v2_base_url}/{client_id}"
        return self._show(url)

    def create(self, payload):
        """
        ``POST /organizations/:organization_name/oauth-clients``

        `OAuth Clients Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/oauth-clients.html#create-an-oauth-client>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/oauth-clients.html#sample-payload>`_
        """
        return self._create(self._org_api_v2_base_url, payload)

    def update(self, client_id, payload):
        """
        ``PATCH /oauth-clients/:id``

        `OAuth Clients Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/oauth-clients.html#update-an-oauth-client>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/oauth-clients.html#sample-payload-1>`_
        """
        url = f"{self._oauth_clients_api_v2_base_url}/{client_id}"
        return self._update(url, payload)

    def destroy(self, client_id):
        """
        ``DELETE /oauth-clients/:id``

        `OAuth Clients Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/oauth-clients.html#destroy-an-oauth-client>`_
        """
        url = f"{self._oauth_clients_api_v2_base_url}/{client_id}"
        return self._destroy(url)



"""
Module for Terraform Cloud API Endpoint: Org Memberships.
"""

from .endpoint import TFCEndpoint
from ._constants import MAX_PAGE_SIZE

class TFCOrgMemberships(TFCEndpoint):
    """
    `Org Memberships API Docs \
        <https://www.terraform.io/docs/cloud/api/organization-memberships.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/organization-memberships"
        self._org_base_url = \
            f"{self._api_v2_base_url}/organizations/{org_name}/organization-memberships"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def invite(self, payload):
        """
        ``POST /organizations/:organization_name/organization-memberships``

        `Org Memberships Invite API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-memberships.html#invite-a-user-to-an-organization>`_

        `Invite Sample Payload \
            <https://www.terraform.io/docs/cloud/api/organization-memberships.html#sample-payload>`_
        """
        return self._create(self._org_base_url, payload)

    def list_for_org(self, query=None, filters=None, page=None, page_size=None):
        """
        ``GET /organizations/:organization_name/organization-memberships``

        `Org Memberships List for Org API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-memberships.html#list-memberships-for-an-organization>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/organization-memberships.html#query-parameters>`_):
            - ``query`` (Optional)
            - ``filter[status]`` (Optional)
            - ``page`` (Optional)
            - ``page_size`` (Optional)

        Example filter(s):

        .. code-block:: python

            filters = [
                {
                    "keys": ["status"],
                    "value": "foo"
                }
            ]

        """
        return self._list(\
            self._org_base_url, query=query, filters=filters, page=page, page_size=page_size)

    def list_all_for_org(self, query=None, filters=None):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every org membership for an organization.

        Returns an array of objects.
        """
        current_page_number = 1
        org_memberships_resp = \
            self._list(self._org_base_url, \
                page=current_page_number, page_size=MAX_PAGE_SIZE, query=query, filters=filters)
        total_pages = org_memberships_resp["meta"]["pagination"]["total-pages"]

        org_memberships = []
        while current_page_number <= total_pages:
            org_memberships_resp = \
                self._list(self._org_base_url, \
                    page=current_page_number, page_size=MAX_PAGE_SIZE, query=query, filters=filters)
            org_memberships += org_memberships_resp["data"]
            current_page_number += 1

        return org_memberships

    def list_for_user(self):
        """
        ``GET /organization-memberships``

        `Org Memberships List for User API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-memberships.html#list-user-39-s-own-memberships>`_
        """
        return self._list(self._endpoint_base_url)

    def show(self, org_membership_id):
        """
        ``GET /organization-memberships/:organization_membership_id``

        `Org Memberships Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-memberships.html#show-a-membership>`_
        """
        url = f"{self._endpoint_base_url}/{org_membership_id}"
        return self._show(url)

    def remove(self, org_membership_id):
        """
        ``DELETE /organization-memberships/:organization_membership_id``

        `Org Memberships Remove API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-memberships.html#remove-user-from-organization>`_
        """
        url = f"{self._endpoint_base_url}/{org_membership_id}"
        return self._destroy(url)
